from django.db.models import Q, QuerySet

from wbfdm.models import Instrument
from wbfdm.sync.abstract import Sync


class MsciInstrumentSync(Sync[Instrument]):
    ESG_PATH = "wbfdm.contrib.msci.dataloaders.esg.MSCIESGDataloader"
    ESG_CONTROVERSY_PATH = "wbfdm.contrib.msci.dataloaders.esg_controversies.MSCIESGControversyDataloader"

    def get_updatable_instruments(self) -> QuerySet[Instrument]:
        return Instrument.objects.filter(
            Q(isin__isnull=False)
            & (Q(dl_parameters__esg__isnull=True) | Q(dl_parameters__esg_controversies__isnull=True))
        )

    def update(self):
        instruments = []

        for instrument in self.get_updatable_instruments():
            self._update_dl_parameters(instrument)
            instruments.append(instrument)

            if len(instruments) % 10000 == 0:
                Instrument.objects.bulk_update(instruments, ["dl_parameters"])
                instruments = []
        Instrument.objects.bulk_update(instruments, ["dl_parameters"])

    def update_or_create_item(self, external_id: int) -> Instrument:
        raise NotImplementedError(f"The {__class__} is not allowed to create instruments")

    def get_item(self, external_id: int) -> dict:
        raise NotImplementedError(f"The {__class__} is not allowed to instantiate from remote resources")

    def _update_dl_parameters(self, instrument) -> Instrument:
        if isin := instrument.isin:
            instrument.dl_parameters["esg"] = {
                "path": self.ESG_PATH,
                "parameters": isin,
            }
            instrument.dl_parameters["esg_controversies"] = {
                "path": self.ESG_CONTROVERSY_PATH,
                "parameters": isin,
            }
        return instrument

    def update_item(self, instrument: Instrument) -> Instrument:
        self._update_dl_parameters(instrument)
        instrument.save()
        return instrument

    def trigger_partial_update(self):
        instruments_to_update = self.get_updatable_instruments()
        instruments_to_update_count = instruments_to_update.count()
        if instruments_to_update_count > 30:
            self.update()
        else:
            for instrument in instruments_to_update:
                self.update_item(instrument)



def register(move_first: bool = False, override_backend: bool = False):
    """
    Decorator to register the metric backend
    """
    from wbfdm.contrib.metric.registry import backend_registry

    def _model_wrapper(backend):
        for key in backend.keys:
            backend_registry.set(
                key, backend.BASKET_MODEL_CLASS, backend, move_first=move_first, override_backend=override_backend
            )
        return backend

    return _model_wrapper



from django.urls import include, path
from wbcore.routers import WBCoreRouter

from . import viewsets

router = WBCoreRouter()

# Representations
router.register(
    r"instrumentmetricrepresentation",
    viewsets.InstrumentMetricRepresentationViewSet,
    basename="instrumentmetricrepresentation",
)

router.register(r"instrumentmetric", viewsets.InstrumentMetricViewSet, basename="instrumentmetric")

urlpatterns = [
    path("", include(router.urls)),
]



from wbcore import serializers as wb_serializers
from wbcore.contrib.geography.serializers import GeographyRepresentationSerializer

from wbfdm.models import Exchange


class ExchangeRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbfdm:exchange-detail")

    class Meta:
        model = Exchange
        fields = ("id", "name", "mic_code", "_detail")


class ExchangeModelSerializer(wb_serializers.ModelSerializer):
    _country = GeographyRepresentationSerializer(source="country")
    _city = GeographyRepresentationSerializer(source="city")

    class Meta:
        model = Exchange
        fields = (
            "id",
            "name",
            "mic_code",
            "operating_mic_code",
            "bbg_exchange_codes",
            "bbg_composite_primary",
            "bbg_composite",
            "refinitiv_identifier_code",
            "refinitiv_mnemonic",
            "country",
            "_country",
            "city",
            "website",
            "opening_time",
            "closing_time",
            "_city",
            "city",
            "comments",
            "apply_round_lot_size",
        )



from collections import defaultdict
from datetime import datetime
from typing import Any, Dict, List, Optional

from django.contrib.auth.models import Permission
from django.db import models
from django.db.models import Q
from slugify import slugify
from wbcore.contrib.authentication.models import User
from wbcore.contrib.io.exceptions import DeserializationError
from wbcore.contrib.io.imports import ImportExportHandler
from wbcore.contrib.notifications.dispatch import send_notification

from .instrument import InstrumentImportHandler


class InstrumentListImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbfdm.InstrumentListThroughModel"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_handler = InstrumentImportHandler(self.import_source)

    def _deserialize(self, data: Dict[str, Any]):
        if from_date := data.get("from_date", None):
            data["from_date"] = datetime.strptime(from_date, "%Y-%m-%d").date()
        if to_date := data.get("to_date", None):
            data["to_date"] = datetime.strptime(to_date, "%Y-%m-%d").date()
        if instrument_list_data := data.pop("instrument_list", None):
            if isinstance(instrument_list_data, int):
                data["instrument_list"] = self.model.instrument_list.get_queryset().get(id=instrument_list_data)
            elif isinstance(instrument_list_data, dict) and "name" in instrument_list_data:
                data["instrument_list"] = self.model.instrument_list.get_queryset().get_or_create(
                    identifier=instrument_list_data.pop("identifier", slugify(instrument_list_data["name"])),
                    **instrument_list_data,
                )[0]
        if instrument_data := data.pop("instrument", None):
            data["instrument"] = self.instrument_handler.process_object(
                instrument_data, only_security=True, read_only=True
            )[0]
        # we try to automatically match the instrument name against a already known matched row
        if instrument_str := data.get("instrument_str"):
            already_existing_rows = self.model.objects.filter(
                instrument__isnull=False,
                instrument_str=instrument_str,
                instrument_list=data.get("instrument_list", None),
            )
            if already_existing_rows.count() == 1:
                data["instrument"] = already_existing_rows.first().instrument
        if "instrument_list" not in data:
            raise DeserializationError("Instrument List not find in this row")

    def _get_instance(self, data: Dict[str, Any], history: Optional[models.QuerySet] = None, **kwargs) -> models.Model:
        if instrument := data.get("instrument", None):
            return self.model.objects.filter(
                instrument=instrument,
                instrument_list=data["instrument_list"],
            ).first()

    def _post_processing_objects(
        self,
        created_objs: List[models.Model],
        modified_objs: List[models.Model],
        unmodified_objs: List[models.Model],
    ):
        objs = modified_objs + unmodified_objs + created_objs
        lists = set(map(lambda x: x.instrument_list, objs))
        leftovers_objs = self.model.objects.filter(instrument_list__in=lists)
        for obj in objs:
            leftovers_objs = leftovers_objs.exclude(id=obj.id)
        leftovers_objs.delete()

        instrument_dict = defaultdict(list)
        for obj in modified_objs + created_objs:
            instrument_dict[obj.instrument_list].append(obj)
        for instrument_list, items in instrument_dict.items():
            if items:
                report = """
                <p>List of instrument added or modified:</p>
                <ul>
                """
                for item in items:
                    if item.instrument:
                        report += f"<li>{item.instrument_str} (automatically link to {item.instrument}</li>"
                    else:
                        report += f"<li>{item.instrument_str}</li>"
                report += "</ul>"
                perm = Permission.objects.get(codename="administrate_instrumentlist")
                for user in (
                    User.objects.filter(is_active=True)
                    .filter(Q(groups__permissions=perm) | Q(user_permissions=perm))
                    .distinct()
                ):
                    send_notification(
                        code="wbfdm.instrument_list_add",
                        name=f"Instruments have been added or modified into the instrument list {instrument_list.name}",
                        body=report,
                        user=user,
                        reverse_name="wbfdm:instrumentlist-detail",
                        reverse_args=[instrument_list.id],
                    )



from .classifications import (
    ClassificationDisplayConfig,
    ClassificationGroupDisplayConfig,
    ClassificationInstrumentRelatedInstrumentDisplayConfig,
    InstrumentClassificationThroughDisplayConfig,
)
from .instrument_lists import InstrumentListDisplayConfig
from .instrument_requests import InstrumentRequestDisplayConfig
from .instruments import InstrumentDisplayConfig
from .instrument_prices import (
    InstrumentPriceDisplayConfig,
    BestAndWorstReturnsInstrumentPandasDisplayConfig,
    FinancialStatisticsInstrumentPandasDisplayConfig,
)
from .instruments_relationships import (
    ClassifiedInstrumentDisplayConfig,
    InstrumentFavoriteGroupDisplayConfig,
    RelatedInstrumentThroughInstrumentDisplayConfig,
)
from .exchanges import ExchangeDisplayConfig
from .monthly_performances import MonthlyPerformancesInstrumentDisplayViewConfig
from .esg import InstrumentESGPAIDisplayViewConfig, InstrumentESGControversyDisplayViewConfig
from .financial_summary import FinancialSummaryDisplayViewConfig



import numpy as np
import pandas as pd

from wbfdm.models import Deal

COLUMNS_MAP = {
    "dealDate": "date",
    "fundedOrg": "equity__provider_id",
    "dealSizeInMillions": "transaction_amount",
    "investors": "investors",
    "fundingRound": "funding_round",
    "fundingRoundCategory": "funding_round_category",
    "valuationInMillions": "valuation",
    "valuationIsEstimate": "valuation_estimated",
    "valuationSourceType": "valuation_source_type",
    "valuationSourceUrls": "valuation_media_mention_source_urls",
    "dealId": "external_id",
}


def parse(import_source):
    def _parse_investors(investors):
        return [{"provider_id": investor["orgId"]} for investor in investors]

    data = []
    df = pd.read_json(import_source.file, orient="records")

    df = df.rename(columns=COLUMNS_MAP)
    df["equity__provider_id"] = df["equity__provider_id"].apply(lambda x: x["orgId"])
    df["type"] = Deal.Types.DEAL
    if not df.empty:
        df["investors"] = df.investors.apply(lambda x: _parse_investors(x))
        df["valuation"] *= 1e6
        df["transaction_amount"] *= 1e6
        df = df.drop(columns=df.columns.difference(COLUMNS_MAP.values()))
        df = df.replace([np.inf, -np.inf, np.nan], None)
        df = df[~df["transaction_amount"].isnull()]
        data.extend(df.to_dict("records"))

    return {"data": data}



from datetime import datetime
from io import BytesIO
from typing import List, Optional

import pandas as pd
from django.db import models
from wbcore.contrib.io.backends import AbstractDataBackend, register

from .mixin import DataBackendMixin
from .utils import Controller

DEFAULT_MAPPING = {
    # Income Statement
    "WC01001": "revenue",
    "WC01051": "cost_of_good_sold_without_depreciation",
    "WC18198": "ebitda",
    "WC18191": "ebit",
    "WC01651": "net_profit",
    "WC08346": "company_tax_rate",
    "WC01201": "cost_research_development",
    "WC01251": "interest_expense",
    "WC01101": "sga",
    # Balance Sheet
    "WC03501": "shareholder_equity",
    "WC02999": "total_assets",
    "WC03101": "current_liabilities",
    "WC03351": "total_liabilities",
    "WC03255": "total_debt",
    "WC02003": "cash_and_cash_equivalents",
    "WC02005": "cash_and_short_term_investments",
    "WC18199": "net_debt",
    "WC02051": "receivables",
    "WC02101": "inventories",
    "WC03040": "payables",
    "WC02201": "current_assets",
    "WC07011": "employee_count",
    "WC18100": "entreprise_value",
    "WC03151": "working_capital",
    "WC05491": "book_value_per_share",
    "WC10010": "eps",
    "WC10030": "diluted_eps",
    "WC01151": "deprecation_and_amortization",
    # Annual data
    "WC04870": "investment_cash",
    "WC04860": "cash_from_operation",
    "WC04890": "financing_cash",
    "WC04601": "capital_expenditures",
    "WC05350": "period__period_end_date",
    "WC05200": "period__period_type",
    # 'WC19109': "net_profit",
    # 'WC19110': "net_profit",
    # 'WC19111': "net_profit",
    # 'WC19112': "net_profit",
}

ANNUAL_FIELDS = ["WC04870", "WC04890", "WC04860", "WC04601"]


@register("Fundamental", provider_key="refinitiv", save_data_in_import_source=False, passive_only=False)
class DataBackend(DataBackendMixin, AbstractDataBackend):
    CHUNK_SIZE = 10
    FISCAL_INTERVAL = 1

    def __init__(self, import_credential: Optional[models.Model] = None, **kwargs):
        self.controller = Controller(import_credential.username, import_credential.password)

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        fields: List[str] = None,
        **kwargs,
    ) -> BytesIO:
        execution_date = execution_time.date()

        if not fields:
            fields = list(DEFAULT_MAPPING.keys())
        if obj_external_ids:
            df_interim = self.controller.get_interim_fundamental_data(
                obj_external_ids,
                fields,
                initial_start=kwargs.get("start", None),
                initial_end=execution_date,
            )
            df_annual = self.controller.get_annual_fundamental_data(
                obj_external_ids,
                fields,
                initial_start=kwargs.get("start", None),
                initial_end=execution_date,
            )
            df = pd.concat([df_interim, df_annual], axis=0)
            df = df.dropna(
                how="all",
                subset=df.columns.difference(["Instrument", "Dates", "period__period_interim", "WC05200", "WC05350"]),
            )
            if not df.empty:
                content_file = BytesIO()
                df.to_json(content_file, orient="records")
                if start := kwargs.get("start", execution_date):
                    file_name = f"fundamental_{start:%Y-%m-%d}_{execution_date:%Y-%m-%d}_{datetime.timestamp(execution_time)}.json"
                else:
                    file_name = f"fundamental_{execution_date:%Y-%m-%d}_{datetime.timestamp(execution_time)}.json"
                yield file_name, content_file



from .llm import run_company_extraction_llm
from .instruments import Instrument, InstrumentType, Cash, Equity
from .instrument_prices import InstrumentPrice
from .instrument_relationships import (
    InstrumentClassificationRelatedInstrument,
    InstrumentClassificationThroughModel,
    InstrumentFavoriteGroup,
    RelatedInstrumentThroughModel,
)
from .classifications import (
    Classification,
    ClassificationGroup,
)
from .instrument_lists import InstrumentList, InstrumentListThroughModel

from .instrument_requests import InstrumentRequest

from .private_equities import Deal
from .options import Option, OptionAggregate



import factory
import rstr

from wbfdm.models import Exchange


class ExchangeFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Exchange
        django_get_or_create = ("mic_code",)

    city = None
    name = factory.LazyAttribute(lambda o: f"Exchange: {o.city}")
    mic_code = factory.LazyAttribute(lambda o: rstr.xeger("([A-Z]{4})"))
    operating_mic_code = factory.LazyAttribute(lambda o: rstr.xeger("([A-Z]{4})"))
    bbg_exchange_codes = factory.LazyAttribute(lambda o: [rstr.xeger("([A-Z]{4})")])
    bbg_composite_primary = factory.Faker("boolean")
    bbg_composite = factory.LazyAttribute(lambda o: rstr.xeger("([A-Z]{4})"))
    refinitiv_identifier_code = factory.LazyAttribute(lambda o: rstr.xeger("([A-Z]{4})"))

    website = factory.Faker("url")
    comments = factory.Faker("paragraph")



from wbcore.metadata.configs.titles import TitleViewConfig

from wbfdm.models import Instrument


class SummaryTableInstrumentChartTitle(TitleViewConfig):
    def get_list_title(self) -> str:
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"{instrument.name_repr} - Financial Summary Table"


class FinancialsGraphInstrumentChartTitle(TitleViewConfig):
    def get_list_title(self) -> str:
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"{instrument.name_repr} - Financials Graph Chart"


class ProfitabilityRatiosInstrumentChartTitle(TitleViewConfig):
    def get_list_title(self) -> str:
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"{instrument.name_repr} - Profitability Ratios Chart"


class CashFlowAnalysisInstrumentTableChartTitle(TitleViewConfig):
    def get_list_title(self) -> str:
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"{instrument.name_repr} - Cash Flow Analysis Table"


class CashFlowAnalysisInstrumentBarChartTitle(TitleViewConfig):
    def get_list_title(self) -> str:
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"{instrument.name_repr} - Cash Flow Analysis Bar Chart"


class NetDebtAndEbitdaInstrumentChartTitle(TitleViewConfig):
    def get_list_title(self) -> str:
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"{instrument.name_repr} - Net debt and EBITDA Chart"


class EarningsInstrumentChartTitle(TitleViewConfig):
    def get_list_title(self) -> str:
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"{instrument.name_repr} - Earnings Analysis Chart"


class FinancialAnalysisGeneratorTitleConfig(TitleViewConfig):
    def get_list_title(self):
        instrument = Instrument.objects.get(id=self.view.kwargs["instrument_id"])
        return f"Ratios for {str(instrument)}"



from wbcore.tests.conftest import *  # isort:skip
from datetime import date

from django.apps import apps
from django.db.models.signals import pre_migrate
from faker import Faker
from pandas.tseries.offsets import BDay
from pytest_factoryboy import register
from wbcore.contrib.authentication.factories import (
    InternalUserFactory,
    SuperUserFactory,
    UserFactory,
)
from wbcore.contrib.currency.factories import CurrencyFactory, CurrencyFXRatesFactory
from wbcore.contrib.directory.factories.entries import (
    CompanyFactory,
    CompanyTypeFactory,
    CustomerStatusFactory,
    EntryFactory,
    PersonFactory,
)
from wbcore.contrib.geography.factories import (
    CityFactory,
    ContinentFactory,
    CountryFactory,
    StateFactory,
)
from wbcore.contrib.geography.tests.signals import app_pre_migration
from wbcore.contrib.io.factories import (
    CrontabScheduleFactory,
    DataBackendFactory,
    ImportSourceFactory,
    ParserHandlerFactory,
    ProviderFactory,
    SourceFactory,
)
from wbfdm.factories import (
    ClassificationFactory,
    ClassificationGroupFactory,
    ExchangeFactory,
    InstrumentClassificationThroughModelFactory,
    InstrumentFactory,
    InstrumentFavoriteGroupFactory,
    InstrumentListFactory,
    InstrumentPriceFactory,
    InstrumentTypeFactory,
    OptionAggregateFactory,
    OptionFactory,
    ParentClassificationFactory,
    RelatedInstrumentThroughModelFactory,
)

fake = Faker()
register(ImportSourceFactory)
register(DataBackendFactory)
register(ProviderFactory)
register(SourceFactory)
register(ParserHandlerFactory)
register(CrontabScheduleFactory)

register(ClassificationFactory)
register(InstrumentClassificationThroughModelFactory)
register(ParentClassificationFactory)
register(ClassificationGroupFactory)
register(ExchangeFactory)
register(InstrumentFactory)
register(InstrumentTypeFactory)
register(InstrumentPriceFactory)
register(InstrumentFavoriteGroupFactory)
register(RelatedInstrumentThroughModelFactory)
register(CurrencyFXRatesFactory)
register(InstrumentListFactory)
register(OptionFactory)
register(OptionAggregateFactory)

register(CurrencyFactory)
register(CityFactory)
register(StateFactory)
register(CountryFactory)
register(ContinentFactory)

register(CompanyFactory)
register(PersonFactory)
register(InternalUserFactory)
register(EntryFactory)
register(CustomerStatusFactory)
register(CompanyTypeFactory)

register(UserFactory)
register(SuperUserFactory, "superuser")

pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("wbfdm"))



from django.contrib import admin

from wbfdm.models import (
    InstrumentClassificationThroughModel,
    InstrumentFavoriteGroup,
    RelatedInstrumentThroughModel,
)


@admin.register(InstrumentFavoriteGroup)
class InstrumentFavoriteGroupModelAdmin(admin.ModelAdmin):
    list_display = ("name", "owner", "public", "primary")
    autocomplete_fields = ["instruments", "owner"]


class InstrumentClassificationThroughModelAdmin(admin.TabularInline):
    list_display = ("classification", "is_favorite")
    model = InstrumentClassificationThroughModel
    fk_name = "instrument"
    autocomplete_fields = ["classification"]


class RelatedInstrumentThroughModelInlineAdmin(admin.TabularInline):
    model = RelatedInstrumentThroughModel
    fk_name = "instrument"
    autocomplete_fields = ["related_instrument"]



from wbcore.metadata.configs.titles import TitleViewConfig

from wbfdm.models import Classification, ClassificationGroup, Instrument


class ClassificationTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Classifications"

    def get_instance_title(self):
        try:
            obj = self.view.get_object()
            return f"Classification: {str(obj)}"
        except AssertionError:
            return "Classification: {{name}}"

    def get_create_title(self):
        return "New Classification"


class ClassificationGroupTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Classification Groups"

    def get_instance_title(self):
        return "Classification Group"

    def get_create_title(self):
        return "New Classification Group"


class ChildClassificationParentClassificationTitleConfig(ClassificationTitleConfig):
    def get_list_title(self):
        classification = Classification.objects.get(id=self.view.kwargs["parent_id"])
        return f"Child Classifications of {classification.name}"

    def get_create_title(self):
        classification = Classification.objects.get(id=self.view.kwargs["parent_id"])
        return f"New Child Classification {classification.name}"


class ClassificationClassificationGroupTitleConfig(ClassificationTitleConfig):
    def get_list_title(self):
        group = ClassificationGroup.objects.get(id=self.view.kwargs["group_id"])
        return f"Entry Level Classifications for group {group.name}"

    def get_create_title(self):
        group = ClassificationGroup.objects.get(id=self.view.kwargs["group_id"])
        return f"New Classification for group {group.name}"


class ClassificationTreeChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        group = ClassificationGroup.objects.get(id=self.view.kwargs["group_id"])
        return f"Classification Tree for group {group.name}"


class ClassificationIcicleChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        group = ClassificationGroup.objects.get(id=self.view.kwargs["group_id"])
        return f"Classification Icicle Chart for group {group.name}"


class InstrumentClassificationThroughTitleConfig(TitleViewConfig):
    def get_create_title(self) -> str:
        if instrument_id := self.view.kwargs.get("instrument_id", None):
            return f"Classification for {Instrument.objects.get(id=instrument_id)}"
        elif classification_id := self.view.kwargs.get("classification_id", None):
            return f"Instrument for {Classification.objects.get(id=classification_id)}"
        else:
            return super().get_create_title()

    def get_instance_title(self):
        try:
            obj = self.view.get_object()
            if obj.classification and obj.instrument:
                return f"Relationship: {str(obj.classification.name)} <-> {str(obj.instrument.name)}"
            return str(obj)
        except AssertionError:
            return super().get_instance_title()



from django.utils.functional import cached_property
from rest_framework.generics import get_object_or_404

from wbfdm.models import Instrument


class InstrumentMixin:
    kwargs: dict

    @cached_property
    def instrument(self) -> Instrument:
        return get_object_or_404(Instrument, pk=self.kwargs["instrument_id"])



from typing import Optional

from django.utils.translation import gettext_lazy as _
from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

from wbfdm.models import InstrumentRequest


class InstrumentRequestDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="instrument_data__name", label="Instrument Name"),
                dp.Field(key="instrument_data__isin", label="Instrument ISIN"),
                dp.Field(key="requester", label="Requester"),
                dp.Field(key="created", label="Created"),
                dp.Field(key="notes", label="Notes"),
                dp.Field(key="handler", label="Handler"),
                dp.Field(key="created_instrument", label="Created Instrument"),
            ],
            legends=[
                dp.Legend(
                    key="status",
                    items=[
                        dp.LegendItem(
                            icon=WBColor.YELLOW_LIGHT.value,
                            label=InstrumentRequest.Status.DRAFT.label,
                            value=InstrumentRequest.Status.DRAFT.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.BLUE_LIGHT.value,
                            label=InstrumentRequest.Status.PENDING.label,
                            value=InstrumentRequest.Status.PENDING.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.GREEN_LIGHT.value,
                            label=InstrumentRequest.Status.APPROVED.label,
                            value=InstrumentRequest.Status.APPROVED.value,
                        ),
                        dp.LegendItem(
                            icon=WBColor.RED_LIGHT.value,
                            label=InstrumentRequest.Status.DENIED.label,
                            value=InstrumentRequest.Status.DENIED.value,
                        ),
                    ],
                ),
            ],
            formatting=[
                dp.Formatting(
                    column="status",
                    formatting_rules=[
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.YELLOW_LIGHT.value},
                            condition=("==", InstrumentRequest.Status.DRAFT.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.BLUE_LIGHT.value},
                            condition=("==", InstrumentRequest.Status.PENDING.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.GREEN_LIGHT.value},
                            condition=("==", InstrumentRequest.Status.APPROVED.value),
                        ),
                        dp.FormattingRule(
                            style={"backgroundColor": WBColor.RED_LIGHT.value},
                            condition=("==", InstrumentRequest.Status.DENIED.value),
                        ),
                    ],
                )
            ],
        )

    def get_instance_display(self) -> Display:
        fields = [
            [repeat_field(3, "status")],
            [repeat_field(3, "notes")],
        ]
        if "pk" in self.view.kwargs:
            fields.append(["requester", "created", "created_instrument"])
        fields.append([repeat_field(3, "data_section")])

        return create_simple_display(
            fields,
            [
                create_simple_section(
                    "data_section",
                    _("Instrument Data"),
                    [
                        [
                            repeat_field(2, "instrument_data__instrument_type"),
                            "instrument_data__name",
                            "instrument_data__name_repr",
                        ],
                        [
                            "instrument_data__isin",
                            "instrument_data__ticker",
                            "instrument_data__refinitiv_identifier_code",
                            "instrument_data__refinitiv_mnemonic_code",
                        ],
                        [repeat_field(2, "instrument_data__currency"), repeat_field(2, "instrument_data__country")],
                        [
                            repeat_field(4, "instrument_data__is_cash"),
                        ],
                        [repeat_field(4, "instrument_data__tags")],
                        [repeat_field(4, "instrument_data__classifications")],
                    ],
                )
            ],
        )



from django.db.models import Q
from import_export import fields
from import_export.widgets import CharWidget, ForeignKeyWidget, ManyToManyWidget
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.geography.import_export.resources.geography import (
    CountryForeignKeyWidget,
)
from wbcore.contrib.io.resources import FilterModelResource

from wbfdm.models import Exchange, Instrument, InstrumentClassificationThroughModel

from .classification import ClassificationManyToManyWidget


class TypeWidget(CharWidget):
    def clean(self, value, row=None, **kwargs):
        return super().clean(value.lower().name(), row=row, **kwargs)


class InstrumentResource(FilterModelResource):
    """
    The resource to download AssetPositions
    """

    instrument_type = fields.Field(column_name="instrument_type", attribute="instrument_type", widget=TypeWidget())

    country = fields.Field(
        column_name="country",
        attribute="country",
        widget=CountryForeignKeyWidget(),
    )

    exchanges = fields.Field(
        column_name="exchanges",
        attribute="exchanges",
        m2m_add=True,
        widget=ManyToManyWidget(Exchange, field="mic_code"),
    )

    currency = fields.Field(
        column_name="currency",
        attribute="currency",
        widget=ForeignKeyWidget(Currency, field="key__iexact"),
    )

    primary_classifications = fields.Field(
        column_name="primary_classifications",
        attribute="classifications",
        m2m_add=True,
        widget=ClassificationManyToManyWidget(primary_classification_group=True),
    )
    default_classifications = fields.Field(
        column_name="default_classifications",
        attribute="classifications",
        m2m_add=True,
        widget=ClassificationManyToManyWidget(primary_classification_group=False),
    )

    def save_m2m(self, obj, data, using_transactions, dry_run):
        """
        override save m2m to define what to to with the nested classification instrument through model values
        """
        if (not using_transactions and dry_run) or self._meta.use_bulk:
            # we don't have transactions and we want to do a dry_run
            # OR use_bulk is enabled (m2m operations are not supported for bulk operations)
            pass
        else:
            for field in self.get_import_fields():
                if isinstance(field.widget, ClassificationManyToManyWidget):
                    defaults = {
                        "is_favorite": str(data.get(f"{field.column_name}__is_favorite", "")).lower()
                        in ["1", "true", "yes"],
                        "reason": data.get(f"{field.column_name}__reason", None),
                        "pure_player": data.get(f"{field.column_name}__pure_player", None),
                        "top_player": data.get(f"{field.column_name}__top_player", None),
                        "percent_of_revenue": data.get(f"{field.column_name}__percent_of_revenue", None),
                    }
                    defaults = {k: v for k, v in defaults.items() if v is not None}
                    for classification in field.clean(data):
                        InstrumentClassificationThroughModel.objects.update_or_create(
                            instrument=obj, classification=classification, defaults=defaults
                        )
                elif isinstance(field.widget, ManyToManyWidget):
                    self.import_field(field, obj, data, True)

    def get_instance(self, instance_loader, row):
        try:
            return Instrument.objects.get(
                Q(isin=row["isin"]) | Q(refinitiv_identifier_code=row["refinitiv_identifier_code"])
            )
        except Instrument.DoesNotExist:
            return None

    class Meta:
        import_id_fields = ("isin",)
        fields = (
            "id",
            "founded_year",
            "inception_date",
            "delisted_date",
            "identifier",
            "name",
            "name_repr",
            "description",
            "isin",
            "ticker",
            "refinitiv_ticker",
            "refinitiv_identifier_code",
            "refinitiv_mnemonic_code",
            "sedol",
            "valoren",
            "headquarter_address",
            "primary_url",
            "is_cash",
            "instrument_type",
            "country",
            "exchanges",
            "currency",
            "primary_classifications",
            "default_classifications",
        )
        export_order = fields
        model = Instrument



# Generated by Django 4.1.9 on 2023-06-14 07:31

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Instrument",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=512)),
            ],
            options={
                "verbose_name": "Instrument",
                "verbose_name_plural": "Instruments",
            },
        ),
        migrations.CreateModel(
            name="InstrumentLookup",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("identifier", models.CharField(max_length=64)),
                ("backend_path", models.CharField(max_length=265)),
                ("prices", models.BooleanField(default=False)),
                ("dividends", models.BooleanField(default=False)),
                ("corporate_actions", models.BooleanField(default=False)),
                ("statements", models.BooleanField(default=False)),
                ("estimates", models.BooleanField(default=False)),
                ("officers", models.BooleanField(default=False)),
                (
                    "instrument",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="lookups", to="wbfdm.instrument"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="InstrumentIdentifier",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("identifier", models.CharField(max_length=32)),
                (
                    "identifier_type",
                    models.CharField(
                        choices=[
                            ("ISIN", "ISIN"),
                            ("RIC", "RIC"),
                            ("BBG", "BBG"),
                            ("TICKER", "Ticker"),
                            ("VALOREN", "Valoren"),
                            ("WKN", "WKN"),
                        ],
                        max_length=32,
                    ),
                ),
                (
                    "instrument",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="identifiers", to="wbfdm.instrument"
                    ),
                ),
            ],
        ),
    ]



import pandas as pd

from wbfdm.enums import CalendarType, Financial, MarketData
from wbfdm.models import Instrument

from .utils import FinancialAnalysisResult, Loader


class StatementWithEstimates:
    """
    Utility class to hold all the potentials statement with estimates table as self contained properties
    """

    def __init__(self, instrument: Instrument, calendar_type: CalendarType = CalendarType.FISCAL):
        self.instrument = instrument
        self.calendar_type = calendar_type

    @property
    def credit_with_estimates(self) -> FinancialAnalysisResult:
        values = [
            Financial.NET_DEBT,
            Financial.EBITDA,
            Financial.TOTAL_DEBT,
            Financial.FREE_CASH_FLOW,
            Financial.CASH_FLOW_FROM_OPERATIONS,
        ]
        loader = Loader(self.instrument, values, self.calendar_type)
        df = loader.load()

        empty_series = pd.Series(dtype="float64", index=df.index)

        df["net_debt_ebitda_ratio"] = df.get(Financial.NET_DEBT.value, empty_series) / df.get(
            Financial.EBITDA.value, empty_series
        )
        df["total_debt_ebitda_ratio"] = df.get(Financial.TOTAL_DEBT.value, empty_series) / df.get(
            Financial.EBITDA.value, empty_series
        )
        df["fcf_total_debt_ratio"] = df.get(Financial.FREE_CASH_FLOW.value, empty_series) / df.get(
            Financial.TOTAL_DEBT.value, empty_series
        )
        df["cfo_total_debt_ratio"] = df.get(Financial.CASH_FLOW_FROM_OPERATIONS.value, empty_series) / df.get(
            Financial.TOTAL_DEBT.value, empty_series
        )

        df.drop(columns=list(map(lambda x: x.value, values)), inplace=True, errors="ignore")

        return FinancialAnalysisResult(df, ignore_group_keys=values, errors=loader.errors)

    @property
    def long_term_solvency_with_estimates(self) -> FinancialAnalysisResult:
        values = [
            Financial.TOTAL_DEBT,
            Financial.SHAREHOLDERS_EQUITY,
            Financial.WORKING_CAPITAL,
        ]
        loader = Loader(self.instrument, values, self.calendar_type)
        df = loader.load()

        empty_series = pd.Series(dtype="float64", index=df.index)

        df["total_debt_equity_ratio"] = df.get(Financial.TOTAL_DEBT.value, empty_series) / df.get(
            Financial.SHAREHOLDERS_EQUITY.value, empty_series
        )
        df["total_debt_capital_ratio"] = df.get(Financial.TOTAL_DEBT.value, empty_series) / df.get(
            Financial.WORKING_CAPITAL.value, empty_series
        )

        df.drop(columns=list(map(lambda x: x.value, values)), inplace=True, errors="ignore")

        return FinancialAnalysisResult(df, ignore_group_keys=values, errors=loader.errors)

    @property
    def short_term_liquidity_with_estimates(self) -> FinancialAnalysisResult:
        values = [
            Financial.CURRENT_ASSETS,
            Financial.CURRENT_LIABILITIES,
            # Financial.INVENTORY, --> Needed for the quick ratio
            Financial.CASH_FLOW_FROM_OPERATIONS,
        ]
        loader = Loader(self.instrument, values, self.calendar_type)
        df = loader.load()

        empty_series = pd.Series(dtype="float64", index=df.index)

        df["current_ratio"] = df.get(Financial.CURRENT_ASSETS.value, empty_series) / df.get(
            Financial.CURRENT_LIABILITIES.value, empty_series
        )
        df["cash_from_operation_current_liabilities_ratio"] = df.get(
            Financial.CASH_FLOW_FROM_OPERATIONS.value, empty_series
        ) / df.get(Financial.CURRENT_LIABILITIES.value, empty_series)
        # df["Cash Flow Margin (%)"] = (df["Inventory"] / df["Cost of goods sold"])  # Inventory does not exist
        df.drop(columns=list(map(lambda x: x.value, values)), inplace=True, errors="ignore")

        return FinancialAnalysisResult(df, ignore_group_keys=values, errors=loader.errors)

    @property
    def asset_turnover_with_estimates(self) -> FinancialAnalysisResult:
        values = [
            Financial.REVENUE,
            Financial.TOTAL_ASSETS,
            Financial.COST_OF_GOODS_SOLD,
            # Financial.INVENTORY,
        ]
        loader = Loader(self.instrument, values, self.calendar_type)
        df = loader.load()

        empty_series = pd.Series(dtype="float64", index=df.index)

        df["asset_turnover"] = df.get(Financial.TOTAL_ASSETS.value, empty_series) / df.get(
            Financial.REVENUE.value, empty_series
        )
        # df["Cash Flow Margin (%)"] = (df["Inventory"] / df["Cost of goods sold"])  # Inventory does not exist
        df.drop(columns=list(map(lambda x: x.value, values)), inplace=True, errors="ignore")

        return FinancialAnalysisResult(df, ignore_group_keys=values, errors=loader.errors)

    @property
    def cashflow_ratios_with_estimates(self) -> FinancialAnalysisResult:
        values = [
            Financial.CASH_FLOW_FROM_OPERATIONS,
            Financial.CURRENT_LIABILITIES,
            Financial.REVENUE,
            Financial.TOTAL_ASSETS,
            Financial.TOTAL_DEBT,
        ]
        loader = Loader(self.instrument, values, self.calendar_type)
        df = loader.load()

        empty_series = pd.Series(dtype="float64", index=df.index)

        df["current_liability_coverage_ratio"] = df.get(
            Financial.CASH_FLOW_FROM_OPERATIONS.value, empty_series
        ) / df.get(Financial.CURRENT_LIABILITIES.value, empty_series)
        df["cash_flow_margin"] = (
            df.get(Financial.CASH_FLOW_FROM_OPERATIONS.value, empty_series)
            / df.get(Financial.REVENUE.value, empty_series)
        ) * 100
        df["asset_efficiency_margin"] = (
            df.get(Financial.CASH_FLOW_FROM_OPERATIONS.value, empty_series)
            / df.get(Financial.TOTAL_ASSETS.value, empty_series)
        ) * 100
        df["net_debt_coverage_ratio"] = df.get(Financial.CASH_FLOW_FROM_OPERATIONS.value, empty_series) / df.get(
            Financial.TOTAL_DEBT.value, empty_series
        )
        df.drop(columns=list(map(lambda x: x.value, values)), inplace=True, errors="ignore")

        return FinancialAnalysisResult(df, ignore_group_keys=values, errors=loader.errors)

    @property
    def margins_with_estimates(self) -> FinancialAnalysisResult:
        values = [
            Financial.REVENUE,
            Financial.GROSS_PROFIT,
            Financial.SGA_EXPENSES,
            Financial.EBITDA,
            Financial.EBIT,
            Financial.NET_INCOME,
            Financial.FREE_CASH_FLOW,
        ]
        loader = Loader(self.instrument, values, self.calendar_type)
        df = loader.load()
        empty_series = pd.Series(dtype="float64", index=df.index)

        df["gross_margin"] = (
            df.get(Financial.GROSS_PROFIT.value, empty_series) / df.get(Financial.REVENUE.value, empty_series)
        ) * 100
        df["sga_margin"] = (
            df.get(Financial.SGA_EXPENSES.value, empty_series) / df.get(Financial.REVENUE.value, empty_series)
        ) * 100
        df["ebitda_margin"] = (
            df.get(Financial.EBITDA.value, empty_series) / df.get(Financial.REVENUE.value, empty_series)
        ) * 100
        df["ebit_margin"] = (
            df.get(Financial.EBIT.value, empty_series) / df.get(Financial.REVENUE.value, empty_series)
        ) * 100
        df["net_income_margin"] = (
            df.get(Financial.NET_INCOME.value, empty_series) / df.get(Financial.REVENUE.value, empty_series)
        ) * 100
        df["free_cash_flow_margin"] = (
            df.get(Financial.FREE_CASH_FLOW.value, empty_series) / df.get(Financial.REVENUE.value, empty_series)
        ) * 100
        df.drop(columns=list(map(lambda x: x.value, values)), inplace=True, errors="ignore")

        return FinancialAnalysisResult(df, ignore_group_keys=values, errors=loader.errors)

    @property
    def summary_with_estimate(self) -> FinancialAnalysisResult:
        values = [
            Financial.REVENUE,
            Financial.GROSS_PROFIT,
            Financial.GROSS_PROFIT_MARGIN,
            Financial.NET_INCOME_REPORTED,
            Financial.EPS_REPORTED,
            Financial.EPS,
            Financial.EBITDA,
            Financial.EBIT,
            Financial.CASH_FLOW_PER_SHARE,
            Financial.BOOK_VALUE_PER_SHARE,
            Financial.CAPEX,
            Financial.SHARES_OUTSTANDING,
            Financial.TOTAL_ASSETS,
            Financial.CURRENT_LIABILITIES,
            Financial.ENTERPRISE_VALUE,
            Financial.CASH_EQUIVALENTS,
        ]
        market_data_values = [
            MarketData.CLOSE,
            MarketData.MARKET_CAPITALIZATION,
        ]
        statement_values = [
            Financial.EMPLOYEES,
            Financial.CASH_AND_SHORT_TERM_INVESTMENT,
            Financial.DILUTED_WEIGHTED_AVG_SHARES,
            Financial.TOTAL_DEBT,
            Financial.NET_DEBT,
            Financial.STOCK_COMPENSATION,
            Financial.TANGIBLE_BOOK_VALUE_PER_SHARE,
        ]
        loader = Loader(
            self.instrument,
            values,
            calendar_type=self.calendar_type,
            market_data_values=market_data_values,
            statement_values=statement_values,
        )
        df = loader.load()
        empty_series = pd.Series(dtype="float64", index=df.index)

        if not df.empty:
            df[Financial.ENTERPRISE_VALUE.value] = (
                df.get(MarketData.MARKET_CAPITALIZATION.value, empty_series)
                + df.get(Financial.NET_DEBT.value, empty_series)
                - df.get(Financial.CASH_EQUIVALENTS.value, empty_series)
            )

            # Calculate a couple of variables
            yearly_df = df.loc[(slice(None), "Y", slice(None), slice(None)), :]
            yearly_empty_series = pd.Series(dtype="float64", index=yearly_df.index)

            quarterly_df = df.loc[
                (
                    slice(None),
                    df.index.isin(["Q1", "Q2", "Q3", "Q4", "S1", "S2", "T1", "T2", "T3"], level=1),
                    slice(None),
                    slice(None),
                ),
                :,
            ].sort_values(by="period_end_date")
            quarterly_empty_series = pd.Series(dtype="float64", index=quarterly_df.index)

            df["price_sales_ratio"] = pd.concat(
                [
                    yearly_df.get(MarketData.MARKET_CAPITALIZATION.value, yearly_empty_series)
                    / yearly_df.get(Financial.REVENUE.value, yearly_empty_series),
                    quarterly_df.get(MarketData.MARKET_CAPITALIZATION.value, quarterly_empty_series)
                    / quarterly_df.get(Financial.REVENUE.value, quarterly_empty_series)
                    .rolling(4, min_periods=1)
                    .sum(),
                ],
                axis=0,
            )
            df["price_earning_reported_ratio"] = pd.concat(
                [
                    yearly_df.get(MarketData.CLOSE.value, yearly_empty_series)
                    / yearly_df.get(Financial.EPS_REPORTED.value, yearly_empty_series),
                    quarterly_df.get(MarketData.CLOSE.value, quarterly_empty_series)
                    / quarterly_df.get(Financial.EPS_REPORTED.value, quarterly_empty_series)
                    .rolling(4, min_periods=1)
                    .sum(),
                ],
                axis=0,
            )
            df["price_earning_ratio"] = pd.concat(
                [
                    yearly_df.get(MarketData.CLOSE.value, yearly_empty_series)
                    / yearly_df.get(Financial.EPS.value, yearly_empty_series),
                    quarterly_df.get(MarketData.CLOSE.value, quarterly_empty_series)
                    / quarterly_df.get(Financial.EPS.value, quarterly_empty_series).rolling(4, min_periods=1).sum(),
                ],
                axis=0,
            )
            df["ev_ebitda_ratio"] = pd.concat(
                [
                    yearly_df.get(Financial.ENTERPRISE_VALUE.value, yearly_empty_series)
                    / yearly_df.get(Financial.EBITDA.value, yearly_empty_series),
                    quarterly_df.get(Financial.ENTERPRISE_VALUE.value, quarterly_empty_series)
                    / quarterly_df.get(Financial.EBITDA.value, quarterly_empty_series).rolling(4, min_periods=1).sum(),
                ],
                axis=0,
            )
            df["ev_ebit_ratio"] = pd.concat(
                [
                    yearly_df.get(Financial.ENTERPRISE_VALUE.value, yearly_empty_series)
                    / yearly_df.get(Financial.EBIT.value, yearly_empty_series),
                    quarterly_df.get(Financial.ENTERPRISE_VALUE.value, quarterly_empty_series)
                    / quarterly_df.get(Financial.EBIT.value, quarterly_empty_series).rolling(4, min_periods=1).sum(),
                ],
                axis=0,
            )

            df["ebitda_margin"] = (
                df.get(Financial.EBITDA.value, empty_series) / df.get(Financial.REVENUE.value, empty_series) * 100
            )
            df["ebit_margin"] = (
                df.get(Financial.EBIT.value, empty_series) / df.get(Financial.REVENUE.value, empty_series) * 100
            )
            df["net_income_margin"] = (
                df.get(Financial.NET_INCOME_REPORTED.value, empty_series)
                / df.get(Financial.REVENUE.value, empty_series)
                * 100
            )
            # df["price_to_cash_flow_ratio"] = df.get(MarketData.CLOSE.value, empty_series) / df.get(Financial.CASH_FLOW_PER_SHARE.value, empty_series)
            df["price_to_cash_flow_ratio"] = pd.concat(
                [
                    yearly_df.get(MarketData.CLOSE.value, yearly_empty_series)
                    / yearly_df.get(Financial.CASH_FLOW_PER_SHARE.value, yearly_empty_series),
                    quarterly_df.get(MarketData.CLOSE.value, quarterly_empty_series)
                    / quarterly_df.get(Financial.CASH_FLOW_PER_SHARE.value, quarterly_empty_series)
                    .rolling(4, min_periods=1)
                    .sum(),
                ],
                axis=0,
            )

            df["price_to_book_ratio"] = df.get(MarketData.CLOSE.value, empty_series) / df.get(
                Financial.BOOK_VALUE_PER_SHARE.value, empty_series
            )
            df["roce"] = pd.concat(
                [
                    yearly_df.get(Financial.EBIT.value, yearly_empty_series)
                    / (
                        yearly_df.get(Financial.TOTAL_ASSETS.value, yearly_empty_series)
                        - yearly_df.get(Financial.CURRENT_LIABILITIES.value, yearly_empty_series)
                    )
                    * 100,
                    quarterly_df.get(Financial.EBIT.value, quarterly_empty_series).rolling(4, min_periods=1).sum()
                    / (
                        quarterly_df.get(Financial.TOTAL_ASSETS.value, quarterly_empty_series)
                        - quarterly_df.get(Financial.CURRENT_LIABILITIES.value, quarterly_empty_series)
                    )
                    * 100,
                ],
                axis=0,
            )

            df["price_to_tangible_bv_ratio"] = df.get(MarketData.CLOSE.value, empty_series) / df.get(
                Financial.TANGIBLE_BOOK_VALUE_PER_SHARE.value, empty_series
            )
            df["cash_shares_ratio"] = df.get(Financial.CASH_AND_SHORT_TERM_INVESTMENT.value, empty_series) / df.get(
                Financial.DILUTED_WEIGHTED_AVG_SHARES.value, empty_series
            )
            df["total_debt_shares_ratio"] = df.get(Financial.TOTAL_DEBT.value, empty_series) / df.get(
                Financial.DILUTED_WEIGHTED_AVG_SHARES.value, empty_series
            )
            df["net_debt_shares_ratio"] = df.get(Financial.NET_DEBT.value, empty_series) / df.get(
                Financial.DILUTED_WEIGHTED_AVG_SHARES.value, empty_series
            )
            df["stock_compensation_employee_ratio"] = df.get(
                Financial.STOCK_COMPENSATION.value, empty_series
            ) / df.get(Financial.EMPLOYEES.value, empty_series)

            df["net_cash"] = df.get(Financial.CASH_EQUIVALENTS.value, empty_series) - df.get(
                Financial.CURRENT_LIABILITIES.value, empty_series
            )
            df[Financial.EBIT.value] = df.get(Financial.EBIT.value, empty_series)

        return FinancialAnalysisResult(
            df,
            ordering=[
                Financial.REVENUE.value,
                "price_sales_ratio",
                Financial.GROSS_PROFIT.value,
                Financial.GROSS_PROFIT_MARGIN.value,
                Financial.NET_INCOME_REPORTED.value,
                "net_income_margin",
                Financial.EPS_REPORTED.value,
                "price_earning_reported_ratio",
                Financial.EPS.value,
                "price_earning_ratio",
                Financial.EBITDA.value,
                "ebitda_margin",
                "ev_ebitda_ratio",
                Financial.EBIT.value,
                "ebit_margin",
                "ev_ebit_ratio",
                Financial.TOTAL_ASSETS.value,
                Financial.CURRENT_LIABILITIES.value,
                "roce",
                Financial.CASH_FLOW_PER_SHARE.value,
                "price_to_cash_flow_ratio",
                Financial.BOOK_VALUE_PER_SHARE.value,
                "price_to_book_ratio",
                Financial.TANGIBLE_BOOK_VALUE_PER_SHARE.value,
                "price_to_tangible_bv_ratio",
                Financial.CASH_AND_SHORT_TERM_INVESTMENT.value,
                Financial.DILUTED_WEIGHTED_AVG_SHARES.value,
                "cash_shares_ratio",
                Financial.TOTAL_DEBT.value,
                "total_debt_shares_ratio",
                Financial.NET_DEBT.value,
                "net_debt_shares_ratio",
                Financial.STOCK_COMPENSATION.value,
                Financial.EMPLOYEES.value,
                "stock_compensation_employee_ratio",
                Financial.CAPEX.value,
                Financial.SHARES_OUTSTANDING.value,
                MarketData.CLOSE.value,
                MarketData.MARKET_CAPITALIZATION.value,
                "net_cash",
            ],
            ignore_group_keys=[
                Financial.TANGIBLE_BOOK_VALUE_PER_SHARE.value,
                Financial.NET_DEBT.value,
                Financial.EMPLOYEES.value,
                Financial.CASH_AND_SHORT_TERM_INVESTMENT.value,
                Financial.DILUTED_WEIGHTED_AVG_SHARES.value,
                Financial.STOCK_COMPENSATION.value,
                Financial.TOTAL_DEBT.value,
            ],
            override_number_with_currency=self.instrument.currency.symbol or self.instrument.currency.key,
            override_number_with_currency_financials=[
                Financial.REVENUE.value,
                Financial.GROSS_PROFIT.value,
                Financial.NET_INCOME_REPORTED.value,
                Financial.EPS.value,
                Financial.EPS_REPORTED.value,
                Financial.EBITDA.value,
                Financial.EBIT.value,
                Financial.TOTAL_ASSETS.value,
                Financial.CURRENT_LIABILITIES.value,
                Financial.TOTAL_DEBT.value,
                Financial.NET_DEBT.value,
                Financial.STOCK_COMPENSATION.value,
                Financial.CAPEX.value,
                MarketData.CLOSE.value,
                MarketData.MARKET_CAPITALIZATION.value,
                Financial.CASH_FLOW_PER_SHARE.value,
                Financial.BOOK_VALUE_PER_SHARE.value,
                Financial.TANGIBLE_BOOK_VALUE_PER_SHARE.value,
                "cash_shares_ratio",
                "total_debt_shares_ratio",
                "net_debt_shares_ratio",
                "stock_compensation_employee_ratio",
                Financial.CASH_AND_SHORT_TERM_INVESTMENT.value,
            ],
            override_number_to_x_financials=[
                "price_sales_ratio",
                "price_earning_reported_ratio",
                "price_earning_ratio",
                "ev_ebitda_ratio",
                "ev_ebit_ratio",
                "price_to_book_ratio",
                "price_to_tangible_bv_ratio",
                "price_to_cash_flow_ratio",
            ],
            override_number_to_percent_financials=[
                Financial.GROSS_PROFIT_MARGIN.value,
                "net_income_margin",
                "ebitda_margin",
                "ebit_margin",
                "roce",
            ],
            errors=loader.errors,
        )

    @property
    def income_statement_with_estimate(self) -> FinancialAnalysisResult:
        values = [
            Financial.REVENUE,
            Financial.COST_OF_GOODS_SOLD,
            Financial.GROSS_PROFIT,
            Financial.GROSS_PROFIT_MARGIN,
            Financial.SELLING_MARKETING_EXPENSES,
            Financial.SGA_EXPENSES,
            Financial.GENERAL_ADMIN_EXPENSES,
            Financial.RND_EXPENSES,
            Financial.STOCK_COMPENSATION,
            Financial.TOTAL_OPERATING_EXPENSES,
            Financial.EBITDA,
            Financial.EBITDA_PER_SHARE,
            Financial.DEPRECATION,
            Financial.AMORTIZATION,
            Financial.EBIT,
            Financial.INTEREST_EXPENSE,
            Financial.NET_INCOME_BEFORE_TAXES,
            Financial.TAX_PROVISION,
            Financial.TAX_RATE,
            Financial.NET_INCOME,
            Financial.SHARES_OUTSTANDING,
            Financial.EPS,
            Financial.EBITDA_REPORTED,
            Financial.NET_INCOME_BEFORE_TAXES_REPORTED,
            Financial.NET_INCOME_REPORTED,
            Financial.EPS_REPORTED,
            Financial.DIVIDEND_PER_SHARE,
        ]
        loader = Loader(self.instrument, values, self.calendar_type)
        df = loader.load()

        return FinancialAnalysisResult(df, ignore_group_keys=values, errors=loader.errors)

    @property
    def balance_sheet_with_estimate(self) -> FinancialAnalysisResult:
        values = [
            Financial.CASH_EQUIVALENTS,
            Financial.INVENTORY,
            Financial.CURRENT_ASSETS,
            Financial.TOTAL_ASSETS,
            Financial.CURRENT_LIABILITIES,
            Financial.CURRENT_DEFERRED_REVENUE,
            Financial.TOTAL_DEBT,
            Financial.NET_DEBT,
            Financial.SHAREHOLDERS_EQUITY,
            Financial.GOODWILL,
            Financial.NET_ASSET_VALUE,
            Financial.BOOK_VALUE_PER_SHARE,
            Financial.TANGIBLE_BOOK_VALUE_PER_SHARE,
            Financial.ENTERPRISE_VALUE,
            Financial.TANGIBLE_BOOK_VALUE,
        ]
        loader = Loader(self.instrument, values, self.calendar_type)
        df = loader.load()

        return FinancialAnalysisResult(df, ignore_group_keys=values, errors=loader.errors)

    @property
    def cash_flow_statement_with_estimate(self) -> FinancialAnalysisResult:
        values = [
            Financial.WORKING_CAPITAL,
            Financial.INCOME_TAXES_PAID,
            Financial.CASH_FLOW_FROM_OPERATIONS,
            Financial.CAPEX,
            Financial.CASH_FLOW_FROM_INVESTING,
            Financial.FREE_CASH_FLOW,
            Financial.FREE_CASH_FLOW_PER_SHARE,
            Financial.TOTAL_DIVIDENDS,
            Financial.CASH_FLOW_FROM_FINANCING,
            Financial.CASH_FLOW_PER_SHARE,
        ]
        loader = Loader(self.instrument, values, self.calendar_type)
        df = loader.load()

        return FinancialAnalysisResult(df, ignore_group_keys=values, errors=loader.errors)

    @property
    def ratios_with_estimate(self) -> FinancialAnalysisResult:
        values = [
            Financial.RETURN_ON_EQUITY,
            Financial.RETURN_ON_INVESTED_CAPITAL,
            Financial.RETURN_ON_CAPITAL,
            Financial.RETURN_ON_ASSETS,
        ]
        loader = Loader(self.instrument, values, self.calendar_type)
        df = loader.load()

        return FinancialAnalysisResult(df, ignore_group_keys=values, errors=loader.errors)



from rest_framework import serializers as rf_serializers
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.serializers import DefaultAttributeFromRemoteField, DefaultFromView

from wbfdm.models.instruments import Classification, ClassificationGroup
from wbfdm.preferences import get_default_classification_group


class ClassificationRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbfdm:classification-detail")

    class Meta:
        model = Classification
        fields = (
            "id",
            "name",
            "computed_str",
            "level",
            "height",
            "code_aggregated",
            "_detail",
        )


class ClassificationZeroHeightRepresentationSerializer(ClassificationRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = {"height": 0}
        if (view := request.parser_context.get("view", None)) and (
            instrument_id := view.kwargs.get("instrument_id", None)
        ):
            filter_params["instruments_neq"] = instrument_id
        return filter_params


class ClassificationIsFavoriteZeroHeightRepresentationSerializer(ClassificationRepresentationSerializer):
    def get_filter_params(self, request):
        filter_params = {
            "height": 0,
            "instruments_through__is_favorite": True,
        }
        if group := get_default_classification_group():
            filter_params["group"] = group.id
        return filter_params


class ClassificationGroupRepresentationSerializer(wb_serializers.RepresentationSerializer):
    max_depth = wb_serializers.IntegerField(read_only=True)
    _detail = wb_serializers.HyperlinkField(reverse_name="wbfdm:classificationgroup-detail")

    class Meta:
        model = ClassificationGroup
        fields = ("id", "name", "is_primary", "max_depth", "_detail")


class ClassificationModelSerializer(wb_serializers.ModelSerializer):
    parent = wb_serializers.PrimaryKeyRelatedField(
        queryset=Classification.objects.all(), default=DefaultAttributeFromRemoteField("parent_id", Classification)
    )
    _parent = ClassificationRepresentationSerializer(source="parent")
    group = wb_serializers.PrimaryKeyRelatedField(
        queryset=ClassificationGroup.objects.all(),
        default=DefaultAttributeFromRemoteField("parent_id", Classification, source="group"),
    )
    _group = ClassificationGroupRepresentationSerializer(source="group")
    code_aggregated = wb_serializers.CharField(required=False, default=DefaultFromView("next_valid_code_aggregated"))
    level_representation = wb_serializers.CharField(required=False, default="")
    description = wb_serializers.TextAreaField(label="Description", allow_null=True, allow_blank=True, required=False)

    @wb_serializers.register_only_instance_resource()
    def additional_resources(self, instance, request, user, **kwargs):
        resources = {
            "childs": reverse("wbfdm:classificationparent-classification-list", args=[instance.id], request=request),
        }
        if instance.children.exists():
            resources["iciclechart"] = (
                f'{reverse("wbfdm:classificationgroup-iciclechart-list", args=[instance.group.id], request=request)}?top_classification={instance.id}'
            )
            resources["treechart"] = (
                f'{reverse("wbfdm:classificationgroup-treechart-list", args=[instance.group.id], request=request)}?top_classification={instance.id}'
            )

        resources["instruments"] = reverse("wbfdm:classification-instrument-list", args=[instance.id], request=request)

        return resources

    def validate(self, data):
        errors = {}
        if parent := data.get("parent", None):
            data["level"] = parent.level + 1
            data["group"] = parent.group
        if (
            (investable := data.get("investable", None))
            and self.instance
            and investable is True
            and self.instance.get_ancestors().filter(investable=False).exists()
        ):
            errors["investable"] = "This classification cannot be investable as long as its parent is non investable."
        if len(errors.keys()) > 0:
            raise rf_serializers.ValidationError(errors)
        return data

    class Meta:
        model = Classification
        fields = (
            "id",
            "parent",
            "_parent",
            "height",
            "group",
            "_group",
            "level",
            "level_representation",
            "name",
            "code_aggregated",
            "investable",
            "description",
            "_additional_resources",
        )


class ClassificationGroupModelSerializer(wb_serializers.ModelSerializer):
    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        if instance.classifications.exists():
            return {
                "classifications": reverse(
                    "wbfdm:classificationgroup-classification-list", args=[instance.id], request=request
                ),
                "treechart": reverse("wbfdm:classificationgroup-treechart-list", args=[instance.id], request=request),
                "iciclechart": reverse(
                    "wbfdm:classificationgroup-iciclechart-list", args=[instance.id], request=request
                ),
            }

        return {}

    class Meta:
        model = ClassificationGroup
        fields = ("id", "name", "is_primary", "max_depth", "code_level_digits", "_additional_resources")



from datetime import date, timedelta
from typing import Iterator

import pypika as pk
from django.db import connections
from pypika import Case
from pypika import functions as fn
from pypika.enums import Order, SqlTypes
from wbcore.contrib.dataloader.dataloaders import Dataloader
from wbcore.contrib.dataloader.utils import dictfetchall

from wbfdm.dataloaders.protocols import FXRateProtocol
from wbfdm.dataloaders.types import FXRateDict


class DatastreamFXRatesDataloader(FXRateProtocol, Dataloader):
    def fx_rates(
        self,
        from_date: date,
        to_date: date,
        target_currency: str,
    ) -> Iterator[FXRateDict]:
        currencies = list(self.entities.values_list("currency__key", flat=True))
        # Define tables
        fx_rate = pk.Table("DS2FxRate")
        fx_code = pk.Table("DS2FxCode")

        # Base query to get data we always need unconditionally
        query = (
            pk.MSSQLQuery.from_(fx_rate)
            # We join on _codes, which removes all instruments not in _codes - implicit where
            .join(fx_code)
            .on(fx_rate.ExRateIntCode == fx_code.ExRateIntCode)
            .where((fx_rate.ExRateDate >= from_date) & (fx_rate.ExRateDate <= to_date + timedelta(days=1)))
            .where(
                (fx_code.ToCurrCode == target_currency)
                & (fx_code.FromCurrCode.isin(currencies))
                & (fx_code.RateTypeCode == "SPOT")
            )
            .orderby(fx_rate.ExRateDate, order=Order.desc)
            .select(
                fn.Cast(fx_rate.ExRateDate, SqlTypes.DATE).as_("fx_date"),
                fn.Concat(fx_code.FromCurrCode, fx_code.ToCurrCode).as_("currency_pair"),
                (Case().when(fx_code.FromCurrCode == target_currency, 1).else_(1 / fx_rate.midrate)).as_("fx_rate"),
            )
        )
        with connections["qa"].cursor() as cursor:
            cursor.execute(query.get_sql())
            yield from dictfetchall(cursor, FXRateDict)



from .classifications import ClassificationButtonConfig, ClassificationGroupButtonConfig
from .instruments import InstrumentButtonViewConfig, ChildrenInstrumentModelViewConfig
from .instrument_prices import (
    InstrumentPriceButtonConfig,
    InstrumentPriceInstrumentButtonConfig,
    FinancialStatisticsInstrumentButtonConfig,
)
from .exchanges import ExchangeButtonConfig



import pandas as pd
from django.contrib import admin
from wbcore.admin import ExportCsvMixin, ImportCsvMixin

from wbfdm.models import Instrument, InstrumentPrice


@admin.register(InstrumentPrice)
class InstrumentPriceAdmin(ExportCsvMixin, ImportCsvMixin, admin.ModelAdmin):
    change_list_template = "wbcore/admin/change_list.html"
    # fsm_field = ["status"]
    list_display = ("date", "instrument", "net_value", "gross_value", "calculated")

    ordering = ("-date",)

    autocomplete_fields = ["instrument"]
    fieldsets = (
        (
            "Main Data",
            {
                "fields": (
                    (
                        "instrument",
                        "date",
                        "import_source",
                        "calculated",
                    ),
                    ("net_value", "gross_value"),
                )
            },
        ),
        (
            "Extra Data",
            {
                "fields": (
                    ("volume", "volume_50d", "volume_200d"),
                    ("outstanding_shares", "market_capitalization", "market_capitalization_consolidated"),
                )
            },
        ),
        (
            "Computed fields",
            {
                "fields": (
                    "lock_statistics",
                    ("sharpe_ratio", "correlation", "beta"),
                )
            },
        ),
    )

    raw_id_fields = ["import_source", "instrument"]

    actions = ["export_as_csv"]

    def manipulate_df(self, df):
        df["instrument"] = df["instrument"].apply(lambda x: Instrument.objects.get(id=x))
        df["date"] = pd.to_datetime(df["date"])
        return df

    def process_model(self, model):
        query_params = {"instrument": model["instrument"], "date": model["date"]}
        self.model.objects.update_or_create(**query_params, defaults=model)



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

ExchangeMenuItem = MenuItem(
    label="Exchange",
    endpoint="wbfdm:exchange-list",
    permission=ItemPermission(
        permissions=["wbfdm.view_exchange"], method=lambda request: is_internal_user(request.user)
    ),
)



from django.utils.translation import gettext_lazy as _
from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.formatting import Condition, Operator

PERFORMANCE_FORMATTING = [
    dp.FormattingRule(
        style={"color": WBColor.GREY.value, "fontWeight": "bold"},
        condition=Condition(Operator("=="), 0),
    ),
    dp.FormattingRule(
        style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
        condition=Condition(Operator("<"), 0),
    ),
    dp.FormattingRule(
        style={"color": WBColor.GREEN_DARK.value, "fontWeight": "bold"},
        condition=Condition(Operator(">"), 0),
    ),
]

BORDER_LEFT = dp.FormattingRule(
    style={
        "borderLeft": "1px solid #bdc3c7",
    }
)


def get_performance_fields(with_comparison_performances: bool = False) -> list[dp.Field]:
    def _get_performance_field(key: str, label: str, show=None):
        if not with_comparison_performances:
            return dp.Field(
                key=f"performance_{key}",
                label=label,
                width=100,
                formatting_rules=PERFORMANCE_FORMATTING,
            )
        return dp.Field(
            key=None,
            label=label,
            width=100,
            show=show,
            children=[
                dp.Field(
                    key=f"performance_{key}",
                    label="Absolute",
                    width=100,
                    formatting_rules=PERFORMANCE_FORMATTING + [BORDER_LEFT],
                ),
                dp.Field(
                    key=f"performance_peer_{key}",
                    label="vs Peer",
                    width=100,
                    formatting_rules=PERFORMANCE_FORMATTING,
                    show="open",
                    tooltip=dp.Tooltip(key="peers"),
                ),
                dp.Field(
                    key=f"performance_benchmark_{key}",
                    label="vs Benchmark",
                    width=100,
                    formatting_rules=PERFORMANCE_FORMATTING,
                    show="open",
                    tooltip=dp.Tooltip(key="benchmarks"),
                ),
            ],
        )

    return [
        dp.Field(
            key=None,
            label="Rolling",
            open_by_default=False,
            children=[
                _get_performance_field(key="daily", label="Daily"),
                _get_performance_field(
                    key="weekly",
                    label="Weekly",
                    show="open",
                ),
                _get_performance_field(
                    key="monthly",
                    label="Monthly",
                    show="open",
                ),
                _get_performance_field(
                    key="quarterly",
                    label="Quarterly",
                    show="open",
                ),
                _get_performance_field(
                    key="yearly",
                    label="Yearly",
                    show="open",
                ),
            ],
        ),
        dp.Field(
            key=None,
            label="Performance-To-Date (PTD)",
            open_by_default=False,
            children=[
                _get_performance_field(key="week_to_date", label="Week", show="open"),
                _get_performance_field(key="month_to_date", label="Month"),
                _get_performance_field(key="quarter_to_date", label="Quarter", show="open"),
                _get_performance_field(key="year_to_date", label="Year", show="open"),
            ],
        ),
        dp.Field(
            key=None,
            label="Previous Performance",
            open_by_default=False,
            show="open",
            children=[
                _get_performance_field(key="previous_week_to_date", label="Week", show="open"),
                _get_performance_field(key="previous_month_to_date", label="Month"),
                _get_performance_field(key="previous_quarter_to_date", label="Quarter", show="open"),
                _get_performance_field(key="previous_year_to_date", label="Year", show="open"),
            ],
        ),
        _get_performance_field(key="inception", label="Inception"),
    ]


def get_statistic_field():
    return dp.Field(
        key=None,
        label=_("Fundamentals"),
        open_by_default=False,
        children=[
            dp.Field(key="statistic_price", label="Last Price", width=100),
            dp.Field(key="statistic_market_capitalization", label="Market Capitalization", width=100),
            dp.Field(key="statistic_revenue_y_1", label="Revenue Y-1", width=100, show="closed"),
            dp.Field(key="statistic_revenue_y0", label="Revenue Y", width=100, show="closed"),
            dp.Field(key="statistic_revenue_y1", label="Revenue Y+1", width=100),
            dp.Field(key="statistic_volume_50d", label="Avg Volume (3m)", width=100, show="closed"),
        ],
    )



import typing

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig

if typing.TYPE_CHECKING:
    from wbfdm.viewsets import StatementPandasViewSet


class StatementDisplayViewConfig(DisplayViewConfig):
    view: "StatementPandasViewSet"
    DEFAULT_COL_WIDTH = 100

    def get_list_display(self) -> dp.ListDisplay:
        def generate_year_field(year: str) -> dp.Field:
            year_col = f"{year}-Y"
            year_column = dp.Field(
                key=year,
                label=year,
                open_by_default=False,
                children=[
                    *map(
                        lambda interim_col: dp.Field(
                            key=interim_col,
                            label=interim_col[5:],
                            show="open",
                            width=self.DEFAULT_COL_WIDTH,
                        ),
                        filter(lambda col: year in col, self.view.interim_columns),
                    ),
                    dp.Field(
                        key=year_col,
                        label="Y",
                        width=self.DEFAULT_COL_WIDTH,
                    ),
                ],
            )

            return year_column

        return dp.ListDisplay(
            fields=[
                dp.Field(key="external_code", label="Code", pinned="left", width=self.DEFAULT_COL_WIDTH),
                dp.Field(key="external_description", label="Description", pinned="left"),
                dp.Field(key="progress", label="Yearly Trend", pinned="left"),
                *map(lambda x: generate_year_field(str(x)), self.view.year_columns),
            ],
        )



from wbcore import filters
from wbcore.filters.defaults import five_year_data_range

from wbfdm.enums import CalendarType, DataType, Indicator, MarketDataChartType
from wbfdm.models.instruments import Instrument


class MarketDataChartFilterSet(filters.FilterSet):
    period = filters.FinancialPerformanceDateRangeFilter(
        label="Period", method="fake_filter", initial=five_year_data_range, required=True
    )
    chart_type = filters.ChoiceFilter(
        method="fake_filter",
        label="Chart Type",
        choices=MarketDataChartType.choices,
        required=True,
        initial=MarketDataChartType.CLOSE,
    )
    benchmarks = filters.ModelMultipleChoiceFilter(
        label="Benchmarks",
        queryset=Instrument.objects.all(),
        endpoint=Instrument.get_representation_endpoint(),
        value_key=Instrument.get_representation_value_key(),
        label_key=Instrument.get_representation_label_key(),
        filter_params={"is_security": True},
        method="fake_filter",
    )
    indicators = filters.MultipleChoiceFilter(
        method="fake_filter",
        label="Indicators",
        choices=Indicator.choices,
        required=False,
    )
    volume = filters.BooleanFilter(
        method="fake_filter",
        label="Add Volume",
        initial=False,
    )
    show_estimates = filters.BooleanFilter(
        method="fake_filter",
        label="Show Estimates",
        initial=True,
    )

    class Meta:
        model = Instrument
        fields = {}


class FinancialRatioFilterSet(filters.FilterSet):
    ttm = filters.BooleanFilter(method="fake_filter", label="TTM/FTM", initial=True, required=True)

    period = filters.FinancialPerformanceDateRangeFilter(
        method="fake_filter", label="Period", initial=five_year_data_range, required=True
    )

    class Meta:
        model = Instrument
        fields = {}


class StatementFilter(filters.FilterSet):
    data_type = filters.ChoiceFilter(
        method="fake_filter",
        label="Data Type",
        choices=DataType.choices,
        required=True,
        initial=DataType.STANDARDIZED,
    )

    class Meta:
        model = Instrument
        fields = {}


class StatementWithEstimateFilter(filters.FilterSet):
    calendar_type = filters.ChoiceFilter(
        method="fake_filter",
        label="Calendar Type",
        choices=CalendarType.choices,
        required=True,
        initial=CalendarType.FISCAL,
    )

    class Meta:
        model = Instrument
        fields = {}



from datetime import date
from typing import TYPE_CHECKING, Optional, Type

from django.contrib.contenttypes.models import ContentType
from django.utils.functional import cached_property
from wbcore import filters
from wbcore.metadata.metadata import WBCoreMetadata
from wbcore.utils.strings import get_aggregate_symbol

from wbfdm.contrib.metric.models import InstrumentMetric
from wbfdm.contrib.metric.registry import backend_registry
from wbfdm.contrib.metric.viewsets.configs.display import (
    InstrumentMetricPivotedListDisplayConfig,
)

from ..dto import MetricKey

if TYPE_CHECKING:
    from django.db.models import Model
    from wbcore.models import WBModel
    from wbcore.viewsets import ModelViewSet

    _Base = ModelViewSet
else:
    _Base = object


class InstrumentMetricMetaData(WBCoreMetadata):
    def determine_metadata(self, request, view):
        metadata = super().determine_metadata(request, view)
        if "filter_fields" in metadata and view.METRIC_SHOW_FILTERS:
            view_metric_key_choices = backend_registry.get_choices(view.METRIC_KEYS)
            metric_show_filter_field = filters.BooleanFilter(
                default=view.METRIC_SHOW_BY_DEFAULT,
                field_name="metric_show_by_default",
                required=True,
                label="Show Metric Keys",
            )
            metric_show_by_default_rep, metric_show_by_default_lookup = metric_show_filter_field.get_representation(
                request, "metric_show_by_default", view
            )
            metadata["filter_fields"]["metric_show_by_default"] = metric_show_by_default_rep
            metadata["filter_fields"]["metric_show_by_default"]["lookup_expr"] = [metric_show_by_default_lookup]

            metric_keys_filter_field = filters.MultipleChoiceFilter(
                choices=view_metric_key_choices,
                field_name="metric_keys",
                required=True,
                label="Metric Keys",
                depends_on=[{"field": "metric_show_by_default", "options": {"activates_on": [True]}}],
                default=list(map(lambda x: x[0], view_metric_key_choices)),
            )
            metric_keys_rep, metric_keys_lookup = metric_keys_filter_field.get_representation(
                request, "metric_keys", view
            )
            metadata["filter_fields"]["metric_keys"] = metric_keys_rep
            metadata["filter_fields"]["metric_keys"]["lookup_expr"] = [metric_keys_lookup]

        return metadata


class InstrumentMetricMixin(_Base):
    """
    Mixin to register automatically a set of metrics defines in METRIC_KEYS into a particular view
    """

    METRIC_KEYS: tuple[MetricKey] | tuple[()] = ()  # The set of MetricKey to inject into the view
    METRIC_BASKET_LABEL: str = "id"  # The filter path of the basket in the queryset
    METRIC_INSTRUMENT_LABEL: str | None = None  # The filter path of the instrument in the queryset. (default to None)
    METRIC_WITH_PREFIXED_KEYS: bool = (
        False  # Set to True if prefixed metrics needs to be appended to the list of metric keys
    )
    METRIC_SHOW_AGGREGATES: bool = True  # set to False if no aggregation needs to be shown for the inserted metrics
    METRIC_SHOW_BY_DEFAULT: bool = True  # If false, the metric are hidden by default
    METRIC_SHOW_FILTERS: bool = False

    display_config_class = InstrumentMetricPivotedListDisplayConfig  # Default display class that automatically regised display Fields into the ListDisplay
    metadata_class = InstrumentMetricMetaData

    @property
    def metric_date(self) -> date | None:
        """
        Property to define at which date the metrics need to be fetched from. Expected to be override
        """
        return None

    @property
    def metric_basket(self) -> Optional["WBModel"]:
        """
        Property to define at for which basket the metrics need to be fetched from. Expected to be override
        """
        return None

    @property
    def metric_basket_class(self) -> Type["Model"]:
        """
        Define the basket class from which the metrics need to be fetched from.

        Default to `metric_basket.__class__` if `metric_basket`  is defined. Otherwise default to the viewset Model property.
        """
        metric_basket_class = self.metric_basket.__class__ if self.metric_basket else self.model
        if not metric_basket_class:
            raise ValueError("Metric Basket Class needs to be defined")
        return metric_basket_class

    @cached_property
    def metric_basket_content_type(self) -> ContentType:
        """
        cached property to store the related basket content type instance
        """
        return ContentType.objects.get_for_model(self.metric_basket_class)

    @property
    def metric_keys(self) -> tuple[MetricKey] | tuple[()]:
        """
        Property used to get the list of metric keys. Can be overridden to define custom logic based on the viewset attributes (e.g. request)
        """
        if self.request.GET.get("metric_show_by_default", "true") == "false":
            return ()
        metric_keys = getattr(self, "METRIC_KEYS", ())
        if metric_keys_filter_repr := self.request.GET.get("metric_keys"):
            metric_keys_filter = metric_keys_filter_repr.split(",")
            metric_keys = tuple(filter(lambda x: x.key in metric_keys_filter, metric_keys))
        return metric_keys

    @property
    def metric_instrument_label(self) -> str | None:
        """
        Property used to get the list of metric instrument label. Can be overridden to define custom logic based on the viewset attributes (e.g. request)
        """
        return getattr(self, "METRIC_INSTRUMENT_LABEL", None)

    @property
    def metric_basket_label(self) -> str:
        """
        Property used to get the list of metric basket label. Can be overridden to define custom logic based on the viewset attributes (e.g. request)
        """
        return getattr(self, "METRIC_BASKET_LABEL", "id")

    @cached_property
    def _metric_serializer_fields(self):
        """
        return the set of serializer fields necessary to display the related metric fields
        """
        extra_serializer_fields = {}
        for metric_key in self.metric_keys:
            metric_backend = backend_registry[metric_key, self.metric_basket_class](self.metric_date)
            extra_serializer_fields.update(
                metric_backend.get_serializer_fields(
                    with_prefixed_key=self.METRIC_WITH_PREFIXED_KEYS, metric_key=metric_key
                )
            )
        return extra_serializer_fields

    def get_ordering_fields(self) -> list[str]:
        """
        Inject the metric field keys as ordering fields
        """
        ordering_fields = list(super().get_ordering_fields())
        for metric_field in self._metric_serializer_fields.keys():
            ordering_fields.append(metric_field)
        return ordering_fields

    def get_queryset(self):
        """
        Annotate the metrics into the viewset queryset
        """
        base_qs = super().get_queryset()
        for metric_key in self.metric_keys:
            base_qs = InstrumentMetric.annotate_with_metrics(
                base_qs,
                metric_key,
                self.metric_basket_class,
                self.metric_date,
                basket_label=self.metric_basket_label,
                instrument_label=self.metric_instrument_label,
            )
        return base_qs

    def get_serializer(self, *args, **kwargs):
        """
        Unwrap defined serializer class and inject the metric fields into a new class
        """

        serializer_class = self.get_serializer_class()
        kwargs.setdefault("context", self.get_serializer_context())

        BaseMeta = serializer_class.Meta
        fields = list(getattr(BaseMeta, "fields", ()))
        read_only_fields = list(getattr(BaseMeta, "read_only_fields", ()))
        for extra_field in self._metric_serializer_fields.keys():
            fields.append(extra_field)
            read_only_fields.append(extra_field)

        Meta = type(str("Meta"), (BaseMeta,), {"fields": fields, "read_only_fields": read_only_fields})
        new_class = type(
            serializer_class.__name__,
            (serializer_class,),
            {
                "Meta": Meta,
                **self._metric_serializer_fields,
                "SERIALIZER_CLASS_FOR_REMOTE_ADDITIONAL_RESOURCES": serializer_class,
            },
        )

        return new_class(*args, **kwargs)

    def get_aggregates(self, queryset, paginated_queryset):
        """
        Automatically register the metric aggregation
        """
        aggregates = dict()
        if self.METRIC_SHOW_AGGREGATES:
            # we try to inject the instrument metric aggregates automatically
            keys_map = {key.key: key for key in self.metric_keys}
            if self.metric_basket and (ct := self.metric_basket_content_type):
                # for each "general" metric (i.e. without particular instrument attached), we add the raw subfield value as aggregates
                for metric in InstrumentMetric.objects.filter(
                    basket_id=self.metric_basket.id,
                    basket_content_type=ct,
                    date=self.metric_date,
                    key__in=keys_map.keys(),
                    instrument__isnull=True,
                ):
                    metric_key = keys_map[metric.key]
                    for subfield_key, subfield_filter in metric_key.subfields_filter_map.items():
                        if subfield_key in self._metric_serializer_fields and (
                            aggregate_fct := metric_key.subfields_map[subfield_key].aggregate
                        ):
                            aggregates[subfield_key] = {
                                get_aggregate_symbol(aggregate_fct.name): metric.metrics.get(subfield_filter)
                            }
            # for all the missings keys (not present in the aggregates already), we compute the aggregatation based on the aggregate function given by the MetricField class
            missing_aggregate_map = {}
            for metric_key in self.metric_keys:
                for field_key, field_filter in metric_key.subfields_filter_map.items():
                    if field_key in self._metric_serializer_fields.keys() and field_key not in aggregates:
                        missing_aggregate_map[field_key] = metric_key.subfields_map[field_key]
            missing_aggregate = queryset.aggregate(
                **{
                    "agg_" + subfield_key: subfield.aggregate(subfield_key)
                    for subfield_key, subfield in missing_aggregate_map.items()
                    if subfield.aggregate is not None
                }
            )
            for k, v in missing_aggregate.items():
                key = k.replace("agg_", "")
                aggregates[key] = {get_aggregate_symbol(missing_aggregate_map[key].aggregate.name): v}
        return aggregates



from contextlib import suppress
from datetime import date
from typing import Any, Generator

from django.db import models
from django.db.utils import DataError
from tqdm import tqdm

from .backends.base import AbstractBackend
from .dto import Metric
from .exceptions import MetricInvalidParameterException
from .models import InstrumentMetric
from .registry import backend_registry


class MetricOrchestrator:
    """
    Orchestrator class responsible to gather the proper metric backends and their respective queryset of baskets

    For each parameters, the orchestrator is responsible to compute and store the generated metrics
    """

    def __init__(self, val_date: date | None, key: str | None = None, basket: Any | None = None, **kwargs):
        self.basket = basket
        self.val_date = val_date
        self.backend_classes = backend_registry.get(metric_key=key, model_class=basket.__class__ if basket else None)

    def _iterate_baskets(self, backend: AbstractBackend) -> models.QuerySet:
        """
        Retrieve and filter the queryset of baskets from the given backend.

        Args:
            backend (AbstractBackend): The backend from which to retrieve the queryset of baskets.

        Returns:
            models.QuerySet: A queryset of baskets, filtered by the current basket if specified.
        """
        qs = backend.get_queryset()
        if self.basket:
            qs = qs.filter(id=self.basket.id)
        return qs

    def _get_parameters(self) -> list[tuple[AbstractBackend, Any]]:
        """
        Generate a list of parameters consisting of metric backend instances and baskets.

        This method initializes backends for each backend class with the validation date,
        iterates through the baskets, and creates a list of tuples containing the backend
        and each basket.

        Returns:
            List[Tuple[AbstractBackend, Any]]: A list of tuples where each tuple contains
            an instance of a instantiated metric backend and a corresponding basket to compute metrics from.
        """
        parameters = []
        for backend_class in self.backend_classes:
            backend = backend_class(self.val_date)
            for basket in self._iterate_baskets(backend):
                parameters.append((backend, basket))
        return parameters

    def get_results(self, debug: bool = False) -> Generator[Metric, None, None]:
        """
        Compute and yield metrics based on the parameters obtained from `_get_parameters`

        Args:
            debug (bool, optional): If True, wraps the parameters list in a tqdm generator for progress tracking. Defaults to False.

        Yields:
            Metric: Each DTO metric computed by the backend for the given baskets.
        """
        parameters = self._get_parameters()
        if debug:
            # if debug mode is enabled, we wrap the parameters list into a tqdm generator
            parameters = tqdm(parameters)
        for parameters in parameters:
            with suppress(MetricInvalidParameterException):
                yield from parameters[0].compute_metrics(parameters[1])

    def process(self, debug: bool = False):
        """
        Process the computation of metrics and update or create InstrumentMetric instances accordingly.

        Side Effects:
            - Updates or creates InstrumentMetric instances based on the computed metrics.
        """
        # we need to see how one threaded loop is fast enough
        errors = []
        for metric in self.get_results(debug=debug):
            try:
                InstrumentMetric.update_or_create_from_metric(metric)
            except DataError:
                errors.append(metric)
        return errors



from django.db.models import TextChoices
from wbcore.utils.enum import ChoiceEnum


class StatementType(ChoiceEnum):
    INCOME_STATEMENT = "incomestatment"
    BALANCE_SHEET = "balancesheet"
    CASHFLOW_STATEMENT = "cashflowstatement"


class MarketData(ChoiceEnum):
    """This enum holds the datatypes about recurring periodic market data, usually daily such as pricing data"""

    OPEN = "open"
    CLOSE = "close"
    HIGH = "high"
    LOW = "low"
    BID = "bid"
    ASK = "ask"
    VWAP = "vwap"
    VOLUME = "volume"
    SHARES_OUTSTANDING = "outstanding_shares"
    MARKET_CAPITALIZATION = "market_capitalization"

    @classmethod
    def name_mapping(cls):
        return {
            cls.CLOSE.value: "Closing Price",
            cls.MARKET_CAPITALIZATION.value: "Market Cap.",
        }


class Frequency(ChoiceEnum):
    """This enum holds the information about the frequency of returned market data"""

    DAILY = "daily"
    WEEKLY = "weekly"
    MONTHLY = "monthly"
    QUARTERLY = "quarterly"
    YEARLY = "yearly"


class Financial(ChoiceEnum):
    """This enum holds the datatypes about recurring financial data, such as revenue"""

    REVENUE = "revenue"
    TOTAL_OPERATING_EXPENSES = "operating_expenses"
    OPERATING_INCOME = "operating_income"
    GROSS_PROFIT = "gross_profit"
    NET_INCOME_BEFORE_TAXES = "pbt"
    NET_INCOME = "net_income"
    EPS = "eps"
    FREE_CASH_FLOW = "free_cash_flow"
    EBITDA = "ebitda"
    EBITDA_PER_SHARE = "ebitda_sh"
    NET_DEBT = "net_debt"
    ENTERPRISE_VALUE = "ev"
    SHARES_OUTSTANDING = "shares_outstanding"
    COST_OF_GOODS_SOLD = "cogs"
    GROSS_PROFIT_MARGIN = "gross_profit_margin"
    SELLING_MARKETING_EXPENSES = "selling_marketing_expenses"
    SGA_EXPENSES = "sga_expenses"
    GENERAL_ADMIN_EXPENSES = "general_admin_expenses"
    RND_EXPENSES = "rnd_expenses"
    STOCK_COMPENSATION = "stock_compensation"
    DEPRECATION = "deprecation"
    AMORTIZATION = "amortization"
    EBIT = "ebit"
    INTEREST_EXPENSE = "interest_expense"
    TAX_PROVISION = "tax_provision"
    TAX_RATE = "tax_rate"
    EBITDA_REPORTED = "ebitda_reported"
    NET_INCOME_BEFORE_TAXES_REPORTED = "pbt_reported"
    NET_INCOME_REPORTED = "net_income_reported"
    EPS_REPORTED = "eps_reported"
    DIVIDEND_PER_SHARE = "dividend_per_share"
    WORKING_CAPITAL = "working_capital"
    INCOME_TAXES_PAID = "income_taxed_paid"
    CASH_FLOW_FROM_OPERATIONS = "cash_flow_from_operations"
    CAPEX = "capex"
    CASH_FLOW_FROM_INVESTING = "cash_flow_from_investing"
    FREE_CASH_FLOW_PER_SHARE = "free_cash_flow_per_share"
    TOTAL_DIVIDENDS = "total_dividends"
    CASH_FLOW_FROM_FINANCING = "cash_flow_from_financing"
    CASH_FLOW_PER_SHARE = "cash_flow_per_share"
    CASH_EQUIVALENTS = "cash_equivalents"
    INVENTORY = "inventory"
    CURRENT_ASSETS = "current_assets"
    TOTAL_ASSETS = "total_assets"
    CURRENT_LIABILITIES = "current_liabilities"
    CURRENT_DEFERRED_REVENUE = "current_deferred_revenue"
    TOTAL_DEBT = "total_debt"
    SHAREHOLDERS_EQUITY = "shareholders_equity"
    GOODWILL = "goodwill"
    NET_ASSET_VALUE = "net_asset_value"
    BOOK_VALUE_PER_SHARE = "book_value_per_share"
    TANGIBLE_BOOK_VALUE_PER_SHARE = "tangible_book_value_per_share"
    TANGIBLE_BOOK_VALUE = "tangible_book_value"
    RETURN_ON_ASSETS = "roa"
    RETURN_ON_CAPITAL = "roc"
    RETURN_ON_INVESTED_CAPITAL = "roic"
    RETURN_ON_EQUITY = "roe"

    EMPLOYEES = "employees"
    CASH_AND_SHORT_TERM_INVESTMENT = "cash_and_short_term_investment"
    DILUTED_WEIGHTED_AVG_SHARES = "diluded_weighted_avg_shares"

    @classmethod
    def name_mapping(cls):
        return {
            cls.REVENUE.value: "Revenue",
            cls.TOTAL_OPERATING_EXPENSES.value: "Total Operating Expenses",
            cls.OPERATING_INCOME.value: "Operating Income",
            cls.GROSS_PROFIT.value: "Gross Profit",
            cls.NET_INCOME_BEFORE_TAXES.value: "Net Income Before Taxes",
            cls.NET_INCOME.value: "Net Income",
            cls.EPS.value: "EPS",
            cls.FREE_CASH_FLOW.value: "Free Cash Flow",
            cls.EBITDA.value: "EBITDA",
            cls.EBITDA_PER_SHARE.value: "EBITDA per Share",
            cls.NET_DEBT.value: "Net Debt",
            cls.ENTERPRISE_VALUE.value: "Enterprise Value",
            cls.SHARES_OUTSTANDING.value: "Shares Outstanding",
            cls.COST_OF_GOODS_SOLD.value: "Cost of goods sold",
            cls.GROSS_PROFIT_MARGIN.value: "Gross Profit Margin",
            cls.SELLING_MARKETING_EXPENSES.value: "Selling/Marketing Expenses",
            cls.SGA_EXPENSES.value: "SG&A Expenses",
            cls.GENERAL_ADMIN_EXPENSES.value: "General & Admin. Expenses",
            cls.RND_EXPENSES.value: "R&D Expenses",
            cls.STOCK_COMPENSATION.value: "Stock based comp.",
            cls.DEPRECATION.value: "Deprecation",
            cls.AMORTIZATION.value: "Amortization",
            cls.EBIT.value: "EBIT",
            cls.INTEREST_EXPENSE.value: "Interest Expenses",
            cls.TAX_PROVISION.value: "Tax Provision",
            cls.TAX_RATE.value: "Tax Rate",
            cls.EBITDA_REPORTED.value: "EBITDA (Reported)",
            cls.NET_INCOME_BEFORE_TAXES_REPORTED.value: "Net Income Before Taxes (Reported)",
            cls.NET_INCOME_REPORTED.value: "Net Income (Reported)",
            cls.EPS_REPORTED.value: "EPS (Reported)",
            cls.DIVIDEND_PER_SHARE.value: "Dividend per Share",
            cls.WORKING_CAPITAL.value: "Working Capital",
            cls.INCOME_TAXES_PAID.value: "Income Taxes Paid",
            cls.CASH_FLOW_FROM_OPERATIONS.value: "Cash Flow from Operations",
            cls.CAPEX.value: "Capital Expenditures",
            cls.CASH_FLOW_FROM_INVESTING.value: "Cash Flow from Investing",
            cls.FREE_CASH_FLOW_PER_SHARE.value: "Free Cash Flow per Share",
            cls.TOTAL_DIVIDENDS.value: "Total Dividends",
            cls.CASH_FLOW_FROM_FINANCING.value: "Cash Flow from Financing",
            cls.CASH_FLOW_PER_SHARE.value: "Cash Flow per Share",
            cls.CASH_EQUIVALENTS.value: "Cash & Cash Equivalents",
            cls.INVENTORY.value: "Inventory",
            cls.CURRENT_ASSETS.value: "Current Assets",
            cls.TOTAL_ASSETS.value: "Total Assets",
            cls.CURRENT_LIABILITIES.value: "Current Liabilities",
            cls.CURRENT_DEFERRED_REVENUE.value: "Current deferred Revenue",
            cls.TOTAL_DEBT.value: "Total Debt",
            cls.SHAREHOLDERS_EQUITY.value: "Shareholders Equity",
            cls.GOODWILL.value: "Goodwill",
            cls.NET_ASSET_VALUE.value: "Net Asset Value",
            cls.BOOK_VALUE_PER_SHARE.value: "Book Value per Share",
            cls.TANGIBLE_BOOK_VALUE_PER_SHARE.value: "Tangible Book Value per Share",
            cls.TANGIBLE_BOOK_VALUE.value: "Tangible Book Value",
            cls.RETURN_ON_ASSETS.value: "Return of Assets (ROA)",
            cls.RETURN_ON_CAPITAL.value: "Return on Captial (ROC)",
            cls.RETURN_ON_INVESTED_CAPITAL.value: "Return on invested Capital (ROIC)",
            cls.RETURN_ON_EQUITY.value: "Return on Equity (ROE)",
            cls.EMPLOYEES.value: "# Employees",
            cls.CASH_AND_SHORT_TERM_INVESTMENT.value: "Cash & Short Term Investments",
            cls.DILUTED_WEIGHTED_AVG_SHARES.value: "Diluted Weighted Average Shares",
            "net_debt_ebitda_ratio": "Net Debt/EBITDA",
            "total_debt_ebitda_ratio": "Total Debt/EBITDA",
            "fcf_total_debt_ratio": "FCF/Total Debt",
            "cfo_total_debt_ratio": "CFO/Total Debt",
            "total_debt_equity_ratio": "Total Debt / Equity",
            "total_debt_capital_ratio": "Total Debt / Capital",
            "current_ratio": "Current Ratio",
            "cash_from_operation_current_liabilities_ratio": "Cash from Ops. To Curr Liab.",
            "asset_turnover": "Asset Turnover",
            "current_liability_coverage_ratio": "Current Liability Coverage Ratio",
            "cash_flow_margin": "Cash Flow Margin (%)",
            "asset_efficiency_margin": "Asset Efficiency (%)",
            "net_debt_coverage_ratio": "Net Debt Coverage Ratio",
            "gross_margin": "Gross Margin (%)",
            "sga_margin": "SG&A Margin (%)",
            "ebitda_margin": "EBITDA Margin (%)",
            "ebit_margin": "EBIT Margin (%)",
            "net_income_margin": "Net Income Margin (%)",
            "free_cash_flow_margin": "Free Cash Flow Margin (%)",
            "price_sales_ratio": "Price / Sales",
            "price_earning_reported_ratio": "P / E(Reported)",
            "price_earning_ratio": "P/E",
            "ev_ebitda_ratio": "EV/EBITDA",
            "ev_ebit_ratio": "EV/EBIT",
            "price_to_cash_flow_ratio": "Price to Cash Flow",
            "price_to_book_ratio": "Price to Book",
            "roce": "ROCE",
            "price_to_tangible_bv_ratio": "Price to Tangible BV",
            "cash_shares_ratio": "Cash/Shares",
            "total_debt_shares_ratio": "Total Debt/Shares",
            "net_debt_shares_ratio": "Net Debt/Shares",
            "stock_compensation_employee_ratio": "Stock based comp. / Employee",
            "net_cash": "Net Cash",
        }


class ESG(ChoiceEnum):
    CARBON_EMISSIONS_SCOPE_1 = "carbon_emissions_scope_1"
    CARBON_EMISSIONS_SCOPE_1_KEY = "carbon_emissions_scope_1_key"
    CARBON_EMISSIONS_SCOPE_2 = "carbon_emissions_scope_2"
    CARBON_EMISSIONS_SCOPE_2_KEY = "carbon_emissions_scope_2_key"
    CARBON_EMISSIONS_SCOPE_3_TOTAL = "carbon_emissions_scope_3_total"
    CARBON_EMISSIONS_SCOPE_3_YEAR = "carbon_emissions_scope_3_year"
    CARBON_EMISSIONS_SOURCE = "carbon_emissions_source"
    CARBON_EMISSIONS_YEAR = "carbon_emissions_year"
    CARBON_EMISSIONS_SCOPE123_EVIC_EUR = "carbon_emissions_scope123_evic_eur"
    CARBON_EMISSIONS_SCOPE123 = "carbon_emissions_scope123"
    CARBON_EMISSIONS_SCOPE123_KEY = "carbon_emissions_scope123_key"
    CARBON_EMISSIONS_SALES_EUR_SCOPE_ALL = "carbon_emissions_sales_eur_scope_all"
    ACTIVE_FF_SECTOR_EXPOSURE = "active_ff_sector_exposure"
    ACTIVE_FF_SECTOR_EXPOSURE_SOURCE = "active_ff_sector_exposure_source"
    ACTIVE_FF_SECTOR_EXPOSURE_YEAR = "active_ff_sector_exposure_year"
    PCT_NON_RENEW_CONSUMPTION_PRODUCTION = "pct_non_renew_consumption_production"
    PCT_TOTAL_NON_RENEW_CONSUMPTION = "pct_total_non_renew_consumption"
    PCT_TOTAL_NON_RENEW_PRODUCTION = "pct_total_non_renew_production"
    TOTAL_RENEW_ENERGY_CONSUMPTION = "total_renew_energy_consumption"
    TOTAL_RENEW_ENERGY_CONSUMPTION_SOURCE = "total_renew_energy_consumption_source"
    TOTAL_RENEW_ENERGY_CONSUMPTION_YEAR = "total_renew_energy_consumption_year"
    TOTAL_NON_RENEW_ENERGY_CONSUMPTION = "total_non_renew_energy_consumption"
    TOTAL_NON_RENEW_ENERGY_CONSUMPTION_SOURCE = "total_non_renew_energy_consumption_source"
    TOTAL_NON_RENEW_ENERGY_CONSUMPTION_YEAR = "total_non_renew_energy_consumption_year"
    TOTAL_NON_RENEW_ENERGY_CONSUMPTION_PRODUCTION = "total_non_renew_energy_consumption_production"
    TOTAL_NON_RENEW_ENERGY_CONSUMPTION_PRODUCTION_SOURCE = "total_non_renew_energy_consumption_production_source"
    TOTAL_NON_RENEW_ENERGY_CONSUMPTION_PRODUCTION_YEAR = "total_non_renew_energy_consumption_production_year"
    TOTAL_ENERGY_CONSUMPTION = "total_energy_consumption"
    TOTAL_ENERGY_CONSUMPTION_SOURCE = "total_energy_consumption_source"
    TOTAL_ENERGY_CONSUMPTION_YEAR = "total_energy_consumption_year"
    ENERGY_CONSUMPTION_INTENSITY_EUR = "energy_consumption_intensity_eur"
    ENERGY_CONSUMPTION_INTENSITY_EUR_SOURCE = "energy_consumption_intensity_eur_source"
    ENERGY_CONSUMPTION_INTENSITY_EUR_YEAR = "energy_consumption_intensity_eur_year"
    OPS_BIODIV_CONTROVERSITIES = "ops_biodiv_controversities"
    OPS_BIODIV_AREAS = "ops_biodiv_areas"
    WATER_EMISSIONS = "water_emissions"
    WATER_EMISSIONS_YEAR = "water_emissions_year"
    WATER_EMISSIONS_SOURCE = "water_emissions_source"
    HAZARD_WASTE = "hazard_waste"
    HAZARD_WASTE_YEAR = "hazard_waste_year"
    HAZARD_WASTE_SOURCE = "hazard_waste_source"
    OECD_ALIGNMENT = "oecd_alignment"
    LABOR_DDP = "labor_ddp"
    COMPLIANCE_GLOBAL_IMPACT = "compliance_global_impact"
    GENDER_PAY_GAP_RATIO = "gender_pay_gap_ratio"
    GENDER_PAY_GAP_RATIO_YEAR = "gender_pay_gap_ratio_year"
    GENDER_PAY_GAP_RATIO_SOURCE = "gender_pay_gap_ratio_source"
    PCT_FEMALE_DIRECTORS = "pct_female_directors"
    CONTROVERSIAL_WEAPONS = "controversial_weapons"
    CONTROVERSIAL_WEAPONS_SOURCE = "controversial_weapons_source"
    CVI_FACTOR = "cvi_factor"
    EVIC_EUR = "evic_eur"
    NACE_SECTION_CODE = "nace_section_code"

    @classmethod
    def mapping(cls):
        return {
            cls.CARBON_EMISSIONS_SCOPE_1.value: (
                1,
                "GHG emissions",
                "Scope 1 GHC Emissions",
                "Carbon Emission Scope 1 (metric tons)",
            ),
            cls.CARBON_EMISSIONS_SCOPE_1_KEY.value: (
                1,
                "GHG emissions",
                "Scope 1 GHC Emissions",
                "Carbon Emissions - Scope 1 KEY",
            ),
            cls.CARBON_EMISSIONS_SCOPE_2.value: (
                1,
                "GHG emissions",
                "Scope 2 GHC Emissions",
                "Carbon Emission Scope 2 (metric tons)",
            ),
            cls.CARBON_EMISSIONS_SCOPE_2_KEY.value: (
                1,
                "GHG emissions",
                "Scope 2 GHC Emissions",
                "Carbon Emissions - Scope 2 KEY",
            ),
            cls.CARBON_EMISSIONS_SCOPE_3_TOTAL.value: (
                1,
                "GHG emissions",
                "Scope 3 Total GHC Emissions",
                "Scope 3 - Total Emissions Estimated",
            ),
            cls.CARBON_EMISSIONS_SCOPE_3_YEAR.value: (
                1,
                "GHG emissions",
                "Scope 3 Total GHC Emissions",
                "Scope 3 - Total Emissions Estimated Year",
            ),
            cls.CARBON_EMISSIONS_SOURCE.value: (
                1,
                "GHG emissions",
                "Total GHG emissions",
                "Carbon Emissions - Source",
            ),
            cls.CARBON_EMISSIONS_YEAR.value: (1, "GHG emissions", "Total GHG emissions", "Carbon Emissions - Year"),
            cls.CARBON_EMISSIONS_SCOPE123.value: (
                2,
                "Carbon footprint",
                "Scope 123 GHC Emissions",
                "Total GHG Emissions (Scopes 1, 2 and 3)",
            ),
            cls.CARBON_EMISSIONS_SCOPE123_KEY.value: (
                2,
                "Carbon footprint",
                "Scope 123 GHC Emissions",
                "Total GHG Emissions (Scopes 1, 2 and 3 Key)",
            ),
            cls.CARBON_EMISSIONS_SCOPE123_EVIC_EUR.value: (
                2,
                "Carbon footprint",
                "Carbon footprint",
                "Total GHG Emissions Intensity per EUR million EVIC (t/EUR million EVIC)",
            ),
            cls.CARBON_EMISSIONS_SALES_EUR_SCOPE_ALL.value: (
                3,
                "GHG intensity of investee companies",
                "GHG intensity of investee companies",
                "Total GHG Emissions Intensity per EUR million Sales (t/EUR million sales)",
            ),
            cls.ACTIVE_FF_SECTOR_EXPOSURE.value: (
                4,
                "Exposure to companies active in the fossil fuel sector",
                "Share of investments in companies active in the fossil fuel sector",
                "Active Fossil Fuel Sector Exposure",
            ),
            cls.ACTIVE_FF_SECTOR_EXPOSURE_SOURCE.value: (
                4,
                "Exposure to companies active in the fossil fuel sector",
                "Share of investments in companies active in the fossil fuel sector",
                "Active Fossil Fuel Sector Exposure - Source",
            ),
            cls.ACTIVE_FF_SECTOR_EXPOSURE_YEAR.value: (
                4,
                "Exposure to companies active in the fossil fuel sector",
                "Share of investments in companies active in the fossil fuel sector",
                "Active Fossil Fuel Sector Exposure - Year",
            ),
            cls.PCT_NON_RENEW_CONSUMPTION_PRODUCTION.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Percentage of non-renewable energy consumption and production",
            ),
            cls.PCT_TOTAL_NON_RENEW_CONSUMPTION.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Percentage of Total Energy Consumption from Non-renewable Sources",
            ),
            cls.PCT_TOTAL_NON_RENEW_PRODUCTION.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Percentage of Total Energy Production from Non-renewable Sources",
            ),
            cls.TOTAL_RENEW_ENERGY_CONSUMPTION.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Total energy consumption from renewable sources (GWh)",
            ),
            cls.TOTAL_RENEW_ENERGY_CONSUMPTION_SOURCE.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Total energy consumption from renewable sources (GWh) Source",
            ),
            cls.TOTAL_RENEW_ENERGY_CONSUMPTION_YEAR.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Total energy consumption from renewable sources (GWh) Year",
            ),
            cls.TOTAL_NON_RENEW_ENERGY_CONSUMPTION.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Total energy consumption from non-renewable sources (GWh)",
            ),
            cls.TOTAL_NON_RENEW_ENERGY_CONSUMPTION_SOURCE.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Total energy consumption from non-renewable sources (GWh) Source",
            ),
            cls.TOTAL_NON_RENEW_ENERGY_CONSUMPTION_YEAR.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Total energy consumption from non-renewable sources (GWh) Year",
            ),
            cls.TOTAL_NON_RENEW_ENERGY_CONSUMPTION_PRODUCTION.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Total Energy Consumption and Production from Non-renewable Sources (GWh)",
            ),
            cls.TOTAL_NON_RENEW_ENERGY_CONSUMPTION_PRODUCTION_SOURCE.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Total Energy Consumption and Production from Non-renewable Sources (GWh) Source",
            ),
            cls.TOTAL_NON_RENEW_ENERGY_CONSUMPTION_PRODUCTION_YEAR.value: (
                5,
                "Share of nonrenewable energy consumption and production",
                "Share of non-renewable energy consumption and non-renewable energy production of investee companies from non-renewable energy sources compared to renewable energy sources, expressed as a percentage",
                "Total Energy Consumption and Production from Non-renewable Sources (GWh) Year",
            ),
            cls.TOTAL_ENERGY_CONSUMPTION.value: (
                5,
                "Energy consumption intensity per high impact climate sector",
                "Total Energy consumption",
                "Total energy consumption (GWh)",
            ),
            cls.TOTAL_ENERGY_CONSUMPTION_SOURCE.value: (
                5,
                "Energy consumption intensity per high impact climate sector",
                "Total Energy consumption",
                "Total energy consumption (GWh) - Source",
            ),
            cls.TOTAL_ENERGY_CONSUMPTION_YEAR.value: (
                5,
                "Energy consumption intensity per high impact climate sector",
                "Total Energy consumption",
                "Total energy consumption (GWh) - Year",
            ),
            cls.ENERGY_CONSUMPTION_INTENSITY_EUR.value: (
                6,
                "Energy consumption intensity per high impact climate sector",
                "Energy consumption in GWh per million EUR of revenue of investee companies, per high impact climate sector",
                "Energy consumption intensity (GWh / EUR million sales)",
            ),
            cls.ENERGY_CONSUMPTION_INTENSITY_EUR_SOURCE.value: (
                6,
                "Energy consumption intensity per high impact climate sector",
                "Energy consumption in GWh per million EUR of revenue of investee companies, per high impact climate sector",
                "Energy consumption intensity (GWh / EUR million sales) - Source",
            ),
            cls.ENERGY_CONSUMPTION_INTENSITY_EUR_YEAR.value: (
                6,
                "Energy consumption intensity per high impact climate sector",
                "Energy consumption in GWh per million EUR of revenue of investee companies, per high impact climate sector",
                "Energy consumption intensity (GWh / EUR million sales) - Year",
            ),
            cls.OPS_BIODIV_CONTROVERSITIES.value: (
                7,
                "Activities negatively affecting biodiversity sensitive areas",
                "Share of investments in investee companies with sites/operations located in or near to biodiversitysensitive areas where activities of those investee companies negatively affect those areas",
                "Company has operations located in biodiversity sensitive areas and is involved in controversies with severe impact on the environment",
            ),
            cls.OPS_BIODIV_AREAS.value: (
                7,
                "Activities negatively affecting biodiversity sensitive areas",
                "Share of investments in investee companies with sites/operations located in or near to biodiversitysensitive areas where activities of those investee companies negatively affect those areas",
                "Operational sites owned, leased, managed in, or adjacent to, protected areas and areas of high biodiversity value outside protected areas",
            ),
            cls.WATER_EMISSIONS.value: (
                8,
                "Emissions to water",
                "Tonnes of emissions to water generated by investee companies per million EUR invested, expressed as a weighted average",
                "Water Emissions (metric tons)",
            ),
            cls.WATER_EMISSIONS_SOURCE.value: (
                8,
                "Emissions to water",
                "Tonnes of emissions to water generated by investee companies per million EUR invested, expressed as a weighted average",
                "Water Emissions (metric tons) - Source",
            ),
            cls.WATER_EMISSIONS_YEAR.value: (
                8,
                "Emissions to water",
                "Tonnes of emissions to water generated by investee companies per million EUR invested, expressed as a weighted average",
                "Water Emissions (metric tons) - Year",
            ),
            cls.HAZARD_WASTE.value: (
                9,
                "Hazardous waste ratio",
                "Tonnes of hazardous waste generated by investee companies per million EUR invested, expressed as a weighted average",
                "Hazardous Waste (metric tons)",
            ),
            cls.HAZARD_WASTE_YEAR.value: (
                9,
                "Hazardous waste ratio",
                "Tonnes of hazardous waste generated by investee companies per million EUR invested, expressed as a weighted average",
                "Hazardous Waste (metric tons) - Year",
            ),
            cls.OECD_ALIGNMENT.value: (
                10,
                "Violations of UN Global Compact principles and Organisation for Economic Cooperation and Development (OECD) Guidelines for Multinational Enterprises",
                "Share of investments in investee companies that have been involved in violations of the UNGC principles or OECD Guidelines for Multinational Enterprises",
                "OECD Alignment",
            ),
            cls.LABOR_DDP.value: (
                11,
                "Lack of processes and compliance mechanisms to monitor compliance with UN Global Compact principles and OECD Guidelines for Multinational Enterprises",
                "Share of investments in investee companies without policies to monitor compliance with the UNGC principles or OECD Guidelines for Multinational Enterprises or grievance /complaints handling mechanisms to address violations of the UNGC principles or OECD Guidelines for Multinational Enterprises",
                "Labor Due Diligence Policy (ILO)",
            ),
            cls.COMPLIANCE_GLOBAL_IMPACT.value: (
                11,
                "Lack of processes and compliance mechanisms to monitor compliance with UN Global Compact principles and OECD Guidelines for Multinational Enterprises",
                "Share of investments in investee companies without policies to monitor compliance with the UNGC principles or OECD Guidelines for Multinational Enterprises or grievance /complaints handling mechanisms to address violations of the UNGC principles or OECD Guidelines for Multinational Enterprises",
                "Mechanism to monitor compliance with UN Global Compact",
            ),
            cls.GENDER_PAY_GAP_RATIO.value: (
                12,
                "Unadjusted gender pay gap",
                "Average unadjusted gender pay gap of investee companies",
                "Gender pay gap ratio",
            ),
            cls.GENDER_PAY_GAP_RATIO_YEAR.value: (
                12,
                "Unadjusted gender pay gap",
                "Average unadjusted gender pay gap of investee companies",
                "Gender pay gap ratio - Year",
            ),
            cls.GENDER_PAY_GAP_RATIO_SOURCE.value: (
                12,
                "Unadjusted gender pay gap",
                "Average unadjusted gender pay gap of investee companies",
                "Gender pay gap ratio - Source",
            ),
            cls.PCT_FEMALE_DIRECTORS.value: (
                13,
                "Board gender diversity",
                "Average ratio of female to male board members in investee companies, expressed as a percentage of all board members",
                "Female Directors Percentage",
            ),
            cls.CONTROVERSIAL_WEAPONS.value: (
                14,
                "Exposure to controversial weapons (antipersonnel mines, cluster munitions, chemical weapons and biological weapons)",
                "Share of investments in investee companies involved in the manufacture or selling of controversial weapons",
                "Exposure to controversial weapons (landmines, cluster munitions, chemical weapons and biological weapons)",
            ),
            cls.CONTROVERSIAL_WEAPONS_SOURCE.value: (
                14,
                "Exposure to controversial weapons (antipersonnel mines, cluster munitions, chemical weapons and biological weapons)",
                "Share of investments in investee companies involved in the manufacture or selling of controversial weapons",
                "Exposure to controversial weapons (landmines, cluster munitions, chemical weapons and biological weapons) - Source",
            ),
            cls.CVI_FACTOR.value: (
                None,
                "Metadata",
                "Capital Value Investor Factor",
                "CVI Factor",
            ),
            cls.EVIC_EUR.value: (
                None,
                "Metadata",
                "Enterprise Value Including Cash",
                "Enterprise Value Including Cash (EUR)",
            ),
            cls.NACE_SECTION_CODE.value: (
                None,
                "Metadata",
                "NACE (Nomenclature of Economic Activities) is the European statistical classification of economic activities.",
                "Nomenclature of Economic Activities (NACE) Code",
            ),
        }


class PeriodType(ChoiceEnum):
    """This enum holds information about the periodicity of the required data"""

    ANNUAL = "annual"
    INTERIM = "interim"
    ALL = "all"


class CalendarType(TextChoices):
    """This enum holds information about the required calendar type"""

    FISCAL = "fiscal"
    CALENDAR = "calendar"


class SeriesType(ChoiceEnum):
    """This enum holds information about the series type"""

    COMPLETE = "complete"
    ACTUAL = "actual"
    ESTIMATE = "estimate"
    FULL_ESTIMATE = "full_estimate"


class DataType(TextChoices):
    """This enum helps to differenciate between reported and standardized data"""

    REPORTED = "reported"
    STANDARDIZED = "standardized"


class EstimateType(ChoiceEnum):
    """This enum helps to differentiate between only valid estimates or all estimates (historic)"""

    ALL = "all"
    VALID = "valid"


class MarketDataChartType(TextChoices):
    CLOSE = ("close", "Close")
    DRAWDOWN = ("drawdown", "Drawdown")
    RETURN = ("ret", "Return")
    LOG_RETURN = ("log-ret", "Log Return")
    CANDLESTICK = ("candlestick", "Candlestick")
    OHLC = ("ohlc", "OHLC")
    # BOLLINGER = ("bollinger", "Bollinger Bands")


class Indicator(TextChoices):
    SMA_50 = ("sma_50", "SMA 50")
    SMA_100 = ("sma_100", "SMA 100")
    SMA_120 = ("sma_120", "SMA 120")
    SMA_200 = ("sma_200", "SMA 200")


class MarketDataSeriesIndicator(ChoiceEnum):
    DRAWDOWN = "drawdown"


class ESGControveryStatus(TextChoices):
    ONGOING = "ONGOING", "Ongoing"
    PARTIALLY_CONCLUDED = "PARTIALLY_CONCLUDED", "Partially Concluded"
    CONCLUDED = "CONCLUDED", "Concluded"


class ESGControverySeverity(TextChoices):
    MINOR = "MINOR", "Minor"
    MODERATE = "MODERATE", "Moderate"
    SEVERE = "SEVERE", "Severe"
    VERY_SEVERE = "VERY_SEVERE", "Very Severe"


class ESGControveryType(TextChoices):
    STRUCTURAL = "STRUCTURAL", "Structural"
    NON_STRUCTURAL = "NON_STRUCTURAL", "Non Structural"


class ESGControveryFlag(TextChoices):
    GREEN = "GREEN", "Green"
    YELLOW = "YELLOW", "Yellow"
    ORANGE = "ORANGE", "Orange"
    RED = "RED", "Red"
    UNKNOWN = "UNKNOWN", "Unknown"



from .instruments import (
    InstrumentModelListSerializer,
    InstrumentModelSerializer,
    InstrumentRepresentationSerializer,
    InstrumentTypeRepresentationSerializer,
    InvestableUniverseRepresentationSerializer,
    CompanyRepresentationSerializer,
    ClassifiableInstrumentRepresentationSerializer,
    SecurityRepresentationSerializer,
    InvestableInstrumentRepresentationSerializer,
    PrimaryInvestableInstrumentRepresentationSerializer,
    EquityRepresentationSerializer,
    ProductRepresentationSerializer,
    ManagedInstrumentRepresentationSerializer,
)
from .instrument_prices import InstrumentPriceModelSerializer, InstrumentPriceInstrumentModelSerializer
from .instrument_relationships import (
    InstrumentClassificationRelatedInstrumentModelSerializer,
    InstrumentClassificationRelatedInstrumentRepresentationSerializer,
    RelatedInstrumentThroughInstrumentModelSerializer,
    InstrumentFavoriteGroupRepresentationSerializer,
    InstrumentFavoriteGroupModelSerializer,
    ReadOnlyInstrumentFavoriteGroupModelSerializer,
    InstrumentClassificationThroughModelSerializer,
)
from .instrument_requests import InstrumentRequestRepresentationSerializer, InstrumentRequestModelSerializer
from .classifications import (
    ClassificationRepresentationSerializer,
    ClassificationZeroHeightRepresentationSerializer,
    ClassificationIsFavoriteZeroHeightRepresentationSerializer,
    ClassificationGroupRepresentationSerializer,
    ClassificationModelSerializer,
    ClassificationGroupModelSerializer,
)
from .instrument_lists import (
    InstrumentListModelSerializer,
    InstrumentListRepresentationSerializer,
    InstrumentListThroughModelSerializer,
)



class QARouter:
    """
    QA from Refinitiv is usually served through an external mssql, oracle or snowflake database.
    We are going to use it in a mostly read-only way, therefore we need a DBRouter that routes
    all database connections for models from QA to the correct database. The database needs to be
    registered in the settings of the project like this:

    ```
    DATABASES = {
        "default": {}, # This is the default django database
        "qa": {},  # This is the important database. The name qa is really important here.
    }
    ```
    Afterwards the settings needs to declare this router explicately as one of the `DATABASE_ROUTERS`
    """

    def db_for_read(self, model, **hints):
        if model._meta.app_label == "qa":
            return "qa"
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if app_label == "qa":
            return False
        return None



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig

from wbfdm.models import InstrumentRequest


class InstrumentRequestEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbfdm:instrumentrequest-list", args=[], request=self.request)

    def get_instance_endpoint(self, **kwargs):
        try:
            obj = self.view.get_object()
            if not self.view.has_validation_permission and obj.status != InstrumentRequest.Status.DRAFT:
                return None
        except AssertionError:
            pass
        return self.get_endpoint(**kwargs)

    def get_delete_endpoint(self, **kwargs):
        return None



# Generated by Django 4.1.9 on 2023-06-23 11:51

import django.db.models.deletion
import wbfdm.models.fk_fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0001_initial"),
    ]

    operations = [
        migrations.RenameField(
            model_name="instrumentlookup",
            old_name="statements",
            new_name="financials",
        ),
        migrations.AddField(
            model_name="instrument",
            name="bbg_lookup",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="wbfdm.instrumentidentifier",
                to_fields={
                    "identifier_type": wbfdm.models.fk_fields.RawFieldValue("BBG"),
                    "instrument": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="financial_backend",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="wbfdm.instrumentlookup",
                to_fields={
                    "financials": wbfdm.models.fk_fields.RawFieldValue(True),
                    "instrument_id": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="isin_lookup",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="wbfdm.instrumentidentifier",
                to_fields={
                    "identifier_type": wbfdm.models.fk_fields.RawFieldValue("ISIN"),
                    "instrument": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="price_backend",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="wbfdm.instrumentlookup",
                to_fields={
                    "instrument_id": wbfdm.models.fk_fields.LocalFieldValue("id"),
                    "prices": wbfdm.models.fk_fields.RawFieldValue(True),
                },
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="ric_lookup",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="wbfdm.instrumentidentifier",
                to_fields={
                    "identifier_type": wbfdm.models.fk_fields.RawFieldValue("RIC"),
                    "instrument": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="ticker_lookup",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="wbfdm.instrumentidentifier",
                to_fields={
                    "identifier_type": wbfdm.models.fk_fields.RawFieldValue("TICKER"),
                    "instrument": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="valoren_lookup",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="wbfdm.instrumentidentifier",
                to_fields={
                    "identifier_type": wbfdm.models.fk_fields.RawFieldValue("VALOREN"),
                    "instrument": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="wkn_lookup",
            field=wbfdm.models.fk_fields.CompositeForeignKey(
                null=True,
                null_if_equal=[],
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="wbfdm.instrumentidentifier",
                to_fields={
                    "identifier_type": wbfdm.models.fk_fields.RawFieldValue("WKN"),
                    "instrument": wbfdm.models.fk_fields.LocalFieldValue("id"),
                },
            ),
        ),
        migrations.AddField(
            model_name="instrumentlookup",
            name="parameters",
            field=models.JSONField(default=dict, null=True),
        ),
    ]



from wbcore import serializers

from wbfdm.enums import (
    ESGControveryFlag,
    ESGControverySeverity,
    ESGControveryStatus,
    ESGControveryType,
)


class InstrumentControversySerializer(serializers.Serializer):
    id = serializers.PrimaryKeyCharField()
    headline = serializers.TextField()
    narrative = serializers.TextField()
    source = serializers.CharField()
    status = serializers.ChoiceField(choices=ESGControveryStatus.choices)
    type = serializers.ChoiceField(choices=ESGControveryType.choices)
    assessment = serializers.ChoiceField(choices=ESGControverySeverity.choices)
    response = serializers.TextField()
    review = serializers.DateField()
    initiated = serializers.DateField()
    flag = serializers.ChoiceField(choices=ESGControveryFlag.choices)

    class Meta:
        fields = (
            "id",
            "headline",
            "narrative",
            "source",
            "status",
            "type",
            "assessment",
            "response",
            "review",
            "initiated",
            "flag",
        )



from datetime import datetime
from typing import List

from django.core.exceptions import ValidationError
from django.db import models
from wbcore.contrib.io.exceptions import DeserializationError
from wbcore.contrib.io.imports import ImportExportHandler

from wbfdm.import_export.handlers.instrument import InstrumentImportHandler

MAX_NB_DATES_BEFORE_HISTORICAL_IMPORT = 3


class InstrumentPriceImportHandler(ImportExportHandler):
    MODEL_APP_LABEL: str = "wbfdm.InstrumentPrice"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instrument_handler = InstrumentImportHandler(self.import_source)
        self.updated_instruments = set()

    def _deserialize(self, data):
        if data.get("instrument", None) is None or data.get("date", None) is None:
            raise DeserializationError("Instrument or Date is empty")
        data["date"] = datetime.strptime(data["date"], "%Y-%m-%d").date()

        instrument = self.instrument_handler.process_object(data["instrument"], only_security=False, read_only=True)[0]
        if not instrument:
            raise DeserializationError("Instrument couldn't be found with given data")
        self.updated_instruments.add(instrument)
        data["instrument"] = instrument
        if data.get("net_value", None) is None:
            # casted_instrument = instrument.get_casted_instrument()
            # # we try to find the primary field among the data to set the net value from it
            # if primary_field_value := data.pop(casted_instrument.primary_field, None):
            #     data["net_value"] = primary_field_value
            # # in a last case, we try to find the close value and set it as the net value
            if close := data.pop("close", None):
                data["net_value"] = close
        else:
            if data.get("gross_value", None) is None:
                data["gross_value"] = data["net_value"]

        # Backward compatibility with refinitiv parser.
        for forbbiden_field in ["close", "price_index", "yield_redemption", "net_return", "offered_rate"]:
            data.pop(forbbiden_field, None)

        # We do this to ensure an invalid number won't make the import fails
        for k, v in data.items():
            if v and (field := self.model._meta.get_field(k)) and hasattr(field, "max_digits"):
                try:
                    field.clean(value=str(round(v, field.decimal_places)), model_instance=None)
                except ValidationError:
                    data[k] = None

        if data.get("net_value", None) is None:
            raise DeserializationError("Net value not set.")

    def _get_instance(self, data, history=None, **kwargs):
        self.import_source.log += f"\nParameter: Instrument={data['instrument']} Date={data['date']}"
        price = data["instrument"].prices.filter(date=data["date"], calculated=False)
        if price.exists():
            return price.first()

    def _save_object(self, _object, **kwargs):
        _object.compute_and_update_statistics()  # compute beta, correlation, sharpe and annualized daily volatility everytime object changes
        return super()._save_object(_object, **kwargs)

    def _post_processing_objects(
        self,
        created_objs: List[models.Model],
        modified_objs: List[models.Model],
        unmodified_objs: List[models.Model],
    ):
        for instrument in self.updated_instruments:
            instrument.update_last_valuation_date()



from django.contrib import admin
from reversion_compare.admin import CompareVersionAdmin

from wbfdm.models import Classification, ClassificationGroup


class ClassificationInLine(admin.TabularInline):
    model = Classification
    fields = ["code_aggregated", "name"]
    extra = 0


@admin.register(ClassificationGroup)
class ClassificationGroupAdmin(CompareVersionAdmin):
    list_display = ["name", "max_depth", "is_primary"]

    search_fields = ["name"]
    ordering = ["name"]

    def reversion_register(self, model, **options):
        options = {
            "ignore_duplicates": True,
        }
        super().reversion_register(model, **options)


@admin.register(Classification)
class ClassificationAdmin(CompareVersionAdmin):
    list_display = ["code_aggregated", "name", "level"]

    search_fields = ["code_aggregated", "name"]
    inlines = [ClassificationInLine]
    ordering = ["code_aggregated"]
    list_filter = ("height", "level", "group")

    def reversion_register(self, model, **options):
        options = {
            "ignore_duplicates": True,
        }
        super().reversion_register(model, **options)



# Generated by Django 5.0.3 on 2024-04-24 12:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0018_instrument_instrument_level_index"),
    ]

    operations = [
        migrations.AlterField(
            model_name="controversy",
            name="source",
            field=models.TextField(blank=True, null=True, verbose_name="Source"),
        ),
    ]



from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers

from wbfdm.models.instruments.instrument_lists import (
    InstrumentList,
    InstrumentListThroughModel,
)
from wbfdm.serializers.instruments import SecurityRepresentationSerializer


class InstrumentListRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbfdm:instrumentlist-detail")

    class Meta:
        model = InstrumentList
        fields = (
            "id",
            "name",
            "_detail",
        )


class InstrumentListModelSerializer(wb_serializers.ModelSerializer):
    @wb_serializers.register_only_instance_resource()
    def instruments(self, instance, request, user, **kwargs):
        if instance:
            base_url = reverse("wbfdm:instrumentlist-instrumentlistthrough-list", args=[instance.id], request=request)
            return {"instruments": base_url}
        return {}

    class Meta:
        model = InstrumentList
        read_only_fields = ("id", "identifier")
        fields = (
            "id",
            "name",
            "identifier",
            "instrument_list_type",
            "_additional_resources",
        )


class InstrumentListThroughModelSerializer(wb_serializers.ModelSerializer):
    _instrument = SecurityRepresentationSerializer(source="instrument")
    _instrument_list = InstrumentListRepresentationSerializer(source="instrument_list")

    class Meta:
        model = InstrumentListThroughModel
        read_only_fields = ("instrument_str",)
        fields = (
            "id",
            "instrument_str",
            "instrument",
            "_instrument",
            "instrument_list",
            "_instrument_list",
            "from_date",
            "to_date",
            "comment",
            "validated",
            "_additional_resources",
        )



import time
from datetime import date, datetime
from typing import Optional

import requests
from celery import shared_task
from tqdm import tqdm


class RateLimitException(Exception):
    pass


class CreditLimitException(Exception):
    pass


class Client:
    MAX_ORG_IDS_SIZE = 100

    def __init__(
        self, client_id: str, client_secret: str, limit: Optional[int] = 10, rate_limit_sleep: Optional[int] = 60
    ):
        self.client_id = client_id
        self.client_secret = client_secret
        self.limit = limit
        self.rate_limit_sleep = rate_limit_sleep
        super().__init__()

    def connect(self):
        token = self._fetch_authorization_token()
        self.jwt_header_value = f"Bearer {token}"

    def _request(self, url, params=None, next_page_token=None):
        if not params:
            params = dict()
        resp = requests.get(
            url,
            params={**params, **{"nextPageToken": next_page_token}} if next_page_token else params,
            headers={"authorization": self.jwt_header_value},
        )
        if resp.status_code != 200:
            if resp.status_code == 429:
                raise RateLimitException()
            raise requests.ConnectionError(
                f"unexpected error from api\nstatus_code: {resp.status_code}\nerror: {resp.text}"
            )

        if credits_remaining_str := resp.headers.get("x-cbinsights-credits-remaining", None):
            if int(credits_remaining_str) <= 0:
                raise CreditLimitException()

        return resp.json()

    def _paginated_request(
        self,
        data_url: str,
        extra_params=None,
        last_update_time: Optional[datetime] = None,
    ):
        params = {"limit": self.limit}
        if last_update_time:
            params["lastUpdateTime"] = last_update_time.isoformat()
        if extra_params:
            params.update(extra_params)
        resp = self._request(data_url, params=params)
        yield resp
        next_page_token = resp["nextPageToken"]
        retry = 0
        while next_page_token and retry < 5:
            try:
                resp = self._request(data_url, params=params, next_page_token=next_page_token)
                yield resp
                next_page_token = resp["nextPageToken"]
            except RateLimitException:
                time.sleep(self.rate_limit_sleep)
                retry += 1
        if retry >= 5:
            raise RateLimitException()

    def _chunk_paginated_request(self, data_url, org_ids, extra_params, endpoint, debug: bool = False):
        data = []
        ranges = list(range(0, len(org_ids), self.MAX_ORG_IDS_SIZE))
        gen = tqdm(ranges, total=len(ranges)) if debug else ranges
        for x in gen:
            ids = org_ids[x : x + 100]
            for res in self._paginated_request(
                data_url,
                extra_params={"orgIds": ",".join(map(str, ids)), **extra_params},
            ):
                if _data := res.get(endpoint, None):
                    data.extend(_data)
        return data

    def _fetch_authorization_token(self):
        auth_url = "https://api.cbinsights.com/v1/authorize"

        auth_resp = requests.get(
            auth_url,
            params={"clientId": self.client_id, "clientSecret": self.client_secret},
        )

        if auth_resp.status_code != 200:
            raise Exception(
                f"unexpected error from api\nstatus_code: {auth_resp.status_code}\nerror: {auth_resp.text}"
            )

        token = auth_resp.json()["token"]

        return token

    def _get_tasks_signatures(self, org_ids, last_update_time, data_url):
        return [
            fetch_datapoint_by_id_asynchronously.s(self.jwt_header_value, data_url.format(org_id), limit=self.limit)
            for org_id in org_ids
        ]

    def fetch_credit_logs(self):
        return self._request("https://api.cbinsights.com/v1/credits")["creditLogs"]

    def fetch_organizations(
        self,
        org_ids: list[str],
        include_datapacks: Optional[str] = "orgSummary,orgKPIs,orgRevenue",
        last_update_time: date = None,
        debug: bool = False,
        **kwargs,
    ):
        extra_params = {"include": include_datapacks, **kwargs}
        if last_update_time:
            extra_params["lastUpdateTime"] = last_update_time.isoformat()
        return self._chunk_paginated_request(
            "https://api.cbinsights.com/v1/organizations", org_ids, extra_params, "organizations", debug=debug
        )

    def fetch_deals(
        self,
        org_ids: list[str],
        start: date = None,
        end: date = None,
        last_update_time: date = None,
        debug: bool = False,
        **extra_params,
    ):
        if start:
            extra_params["minDealDate"] = start.isoformat()
        if end:
            extra_params["maxDealDate"] = end.isoformat()
        if last_update_time:
            extra_params["lastUpdateTime"] = last_update_time.isoformat()
        return self._chunk_paginated_request(
            "https://api.cbinsights.com/v1/deals", org_ids, extra_params, "deals", debug=debug
        )

    # Sub deals utility functions.
    def fetch_fundings_for_id(self, org_id: str):
        data = []
        for res in self._paginated_request(f"https://api.cbinsights.com/v1/organizations/{org_id}/fundings"):
            if fundings_data := res.get("fundings", None):
                data.extend(fundings_data)
        return data

    def fetch_investments_for_id(self, org_id: str):
        data = []
        for res in self._paginated_request(f"https://api.cbinsights.com/v1/organizations/{org_id}/investments"):
            if investments_data := res.get("investments", None):
                data.extend(investments_data)
        return data

    def fetch_portfolioexits_for_id(self, org_id: str):
        data = []
        for res in self._paginated_request(f"https://api.cbinsights.com/v1/organizations/{org_id}/portfolioExits"):
            if portfolioexits_data := res.get("portfolioExits", None):
                data.extend(portfolioexits_data)
        return data

    def fetch_org_id(self, org_name, org_urls: list[str]) -> int:
        org_id = None
        i = 0
        while not org_id and i < len(org_urls):
            params = {"url": org_urls[i], "orgName": org_name}
            try:
                hits = self._request(
                    "https://api.cbinsights.com/v1/organizations/lookup",
                    params={k: v for k, v in params.items() if v},
                )["hits"]
                if len(hits) > 0:
                    org_id = hits[0]["orgId"]
            except requests.ConnectionError as e:
                print(e)  # noqa: T201
            i += 1
        return org_id


@shared_task(
    queue="importexport",
    autoretry_for=(Exception,),
    exponential_backoff=2,
    retry_kwargs={"max_retries": 2},
    retry_jitter=False,
)
def fetch_datapoint_by_id_asynchronously(
    jwt_header_value,
    data_url,
    last_update_time: Optional[datetime] = None,
    limit: Optional[int] = 10,
    rate_limit_sleep: Optional[int] = 60,
):
    params = {"limit": limit}
    if last_update_time:
        params["lastUpdateTime"] = last_update_time.isoformat()
    data = []

    resp = Client._request(jwt_header_value, data_url, params=params)
    data.append(resp)

    return data



from wbfdm.import_export.backends.refinitiv.fundamental import DEFAULT_MAPPING

from .utils import parse_periodic_fundamental_data


def parse(import_source):
    return parse_periodic_fundamental_data(
        import_source, DEFAULT_MAPPING, extra_normalization_map={"company_tax_rate": 0.01}
    )
